/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "VMDragDrop.h"
#include "SecureFuncs.h"
#include "WinTools.h"
#include "DbgPrintf.h"
#include "DllMain.h"
#include "HostCalls.h"
#include "GServClient.h"
#include "CriticalSection.h"
#include <map>
#include "Settings.h"
using namespace std;


typedef struct _FINDVIRTBOX_PARAMS
{
	DWORD dwProcessId;
	HWND  hWnd;
	BOOL  bWasError;

}FINDVIRTBOX_PARAMS;


BOOL
CALLBACK
DragDropGetCopiedFileListHandler(
    UINT DropHandle,
    vector<wstring> *pClipboardFileList
    );



UINT
DragDrop_CreateHandle( HDROP hDrop );

void
VM_EnableDragDrop( HWND hWndOut );

VOID
OnDropFilesMsg( HDROP hDrop );



HANDLE g_VMDragDropInitThread = NULL;
HWND g_hWndVboxOut = NULL;
volatile BOOL g_Quiting = FALSE;
WNDPROC g_OrigOutputWindowProc = NULL;
HANDLE g_SubClassWndEvent = NULL;

map<UINT, vector<wstring> > g_DroppedFilesList;
CCriticalSection g_DroppedFilesListLock;
volatile LONG g_DroppedFilesListLastHandle = 1;


BOOL
CALLBACK
FindVirtBox_EnumChildProc(
  HWND hwnd,
  LPARAM lParam
  )
{
	FINDVIRTBOX_PARAMS *pParams = (FINDVIRTBOX_PARAMS *)lParam;
	//BOOL bRet = FALSE;

	/*CHAR szClassName[100] = "";

	bRet = GetClassName( hwnd, szClassName, ARRAYSIZE(szClassName) );

	if ( !bRet )
	{
		return TRUE;
	}*/

	//CHAR szWindowText[200] = "";

	//GetWindowText( hwnd, szWindowText, ARRAYSIZE(szWindowText) );

	RECT rect;
	RtlZeroMemory( &rect, sizeof(rect) );
	GetClientRect( hwnd, &rect);

	RECT rectParent;
	RtlZeroMemory( &rectParent, sizeof(rectParent) );

    //DbgPrintf( "Child Window: 0x%p\n", hwnd );

    HWND hwndParent = NULL;
    HWND hwndParentParent = NULL;

	if ( rect.bottom > 0 && rect.right > 0 && 
         FindFirstChildWindow(hwnd) == NULL &&
         ( hwndParent = GetParent( hwnd ) ) != NULL &&
         ( hwndParentParent = GetParent( hwndParent ) ) != NULL &&
         /*GetParent( hwndParentParent ) != NULL &&*/
         GetClientRect( hwndParent, &rectParent )
         )
	{
        //DbgPrintf( "Check Window: 0x%p (%u,%u,%u,%u)\n", hwnd, (ULONG)rect.left, (ULONG)rect.top, (ULONG)rect.right, (ULONG)rect.bottom );
        //DbgPrintf( "Parent Window: 0x%p (%u,%u,%u,%u)\n", hwndParent, (ULONG)rectParent.left, (ULONG)rectParent.top, (ULONG)rectParent.right, (ULONG)rectParent.bottom );

        if ( rectParent.right == rect.right && rectParent.bottom == rect.bottom )
        {
			if ( pParams->hWnd && pParams->hWnd != hwnd )
			{
				pParams->bWasError = TRUE; //several wnds were found
                DbgPrintf( "Error: 2 candidates found for VBOX Output(%p,%p)\n", hwnd, pParams->hWnd );
			}
			else
			{
                //DbgPrintf( "Found!!!: %p\n", hwnd );
				pParams->hWnd = hwnd;
			}
        }
	}

    

    /*EnumChildWindows(
		hwnd,
		FindVirtBox_EnumChildProc,
		lParam
		);*/

	return TRUE;
}



BOOL 
CALLBACK 
FindVirtBox_EnumWindowsProc(          
	HWND hwnd,
    LPARAM lParam
	)
{
	FINDVIRTBOX_PARAMS *pParams = (FINDVIRTBOX_PARAMS *)lParam;
	DWORD dwWndPid = 0;

	GetWindowThreadProcessId(
		hwnd,
		&dwWndPid
		);

	if ( dwWndPid != pParams->dwProcessId )
		return TRUE;

	//CHAR szClassName[100] = "";

	//bRet = GetClassNameA( hwnd, szClassName, ARRAYSIZE(szClassName) );

	//RECT rect;
	//RtlZeroMemory( &rect, sizeof(rect) );
	//GetClientRect( hwnd, &rect);

    //DbgPrintf( "Parent Window: 0x%p\n", hwnd );

    EnumChildWindows(
		hwnd,
		FindVirtBox_EnumChildProc,
		lParam
		);

	//continue search in case to find 2+ appropriate windows

	return TRUE;
}


HWND
FindVirtBoxOutputWindow( ULONG ProcessId )
{
	FINDVIRTBOX_PARAMS Params;

	RtlZeroMemory( &Params, sizeof(Params) );

	Params.dwProcessId = ProcessId;

	EnumWindows( FindVirtBox_EnumWindowsProc, (LPARAM)&Params );

	if ( Params.bWasError )
		return NULL;

	return Params.hWnd;
}


LRESULT 
CALLBACK 
OutputWindowProcHook(
  HWND hwnd,
  UINT uMsg,
  WPARAM wParam,
  LPARAM lParam
  )
{

    if ( WM_DROPFILES == uMsg )
    {        
        OnDropFilesMsg( (HDROP)wParam );
        return 0; //proceed this message
    }


    else if ( WM_DESTROY == uMsg )
    {
        DbgPrintf( "OutputWindowProcHook: Destroying window: %p\n", hwnd );
        SetEvent( g_SubClassWndEvent );

    }

    return CallWindowProc( g_OrigOutputWindowProc, hwnd, uMsg, wParam, lParam );
}


DWORD 
WINAPI 
VMDragDropInitThread(
    LPVOID lpParameter
    )
{
    HWND hWndOut = NULL;
    DWORD dwProcessId = GetCurrentProcessId();    

    g_SubClassWndEvent = CreateEvent( 
        NULL, 
        FALSE, //bManualReset
        FALSE, //bInitialState
        NULL
        );


    if ( NULL == g_SubClassWndEvent )
    {
        DbgPrintf( "CreateEvent failed\n" );
        CloseHandle( g_VMDragDropInitThread );
        g_VMDragDropInitThread = NULL;
        return -1;
    }

    BOOL bWasError = FALSE;

    do
    {
        while ( !g_Quiting )
        {
            hWndOut = FindVirtBoxOutputWindow(
                dwProcessId
                );

            if ( hWndOut && 
                 (PVOID)GetWindowLongPtr( hWndOut, GWLP_WNDPROC ) !=  OutputWindowProcHook
                )
            {
                //be sure that it's not old window
                break;
            }

            Sleep( 100 );
        }

        if ( NULL == hWndOut || g_Quiting )
            break;

        VM_EnableDragDrop( hWndOut );

        DWORD dwWaitCode = 0;

        do 
        {
            dwWaitCode = WaitForSingleObject( g_SubClassWndEvent, 200 );

            if ( dwWaitCode == WAIT_OBJECT_0 )
            {
                break;
            }
            else if ( dwWaitCode != WAIT_TIMEOUT )
            {
                bWasError = TRUE;
                DbgPrintf( "Error: dwWaitCode != WAIT_TIMEOUT\n" );
                break;
            }

        } while ( dwWaitCode == WAIT_TIMEOUT && !g_Quiting );
        
    } while ( !g_Quiting && !bWasError );
    

    if ( g_SubClassWndEvent )
    {
        CloseHandle( g_SubClassWndEvent );
        g_SubClassWndEvent = NULL;
    }

    CloseHandle( g_VMDragDropInitThread );
    g_VMDragDropInitThread = NULL;
    return 0;
}


VOID
VMDragDropInit()
{
    DbgPrintf( "VMDragDropInit\n" );
	
	if ( !Settings_DragDropEnabled() )
	{
		DbgPrintf( "DragDrop disabled\n" );
		return;
	}

    SetGetCopiedFileListHandler( DragDropGetCopiedFileListHandler );

    DWORD dwTid = 0;

    g_VMDragDropInitThread = CreateThread(
        NULL,
        0,
        (LPTHREAD_START_ROUTINE)VMDragDropInitThread,
        NULL,
        0,
        &dwTid
        );

    if ( NULL == g_VMDragDropInitThread )
    {
        DbgPrintf( "Cannot create VMDragDropInitThread\n" );
    }

}

VOID
VMDragDropClear()
{
    DbgPrintf( "VMDragDropClear\n" );

    g_Quiting = TRUE;

    if ( g_VMDragDropInitThread )
    {
        if ( g_SubClassWndEvent )
        {
            DbgPrintf( "Setting g_SubClassWndEvent\n" );
            SetEvent( g_SubClassWndEvent );
        }

        DbgPrintf( "WaitForSingleObject( g_VMDragDropInitThread\n" );

        DWORD dwWaitCode = WaitForSingleObject( g_VMDragDropInitThread, 500 );

        if ( dwWaitCode == WAIT_TIMEOUT )
        {
            DbgPrintf( "Error: killing thread VMDragDropInitThread\n" );
            TerminateThread( g_VMDragDropInitThread, 0 );
            g_VMDragDropInitThread = NULL;

            if ( g_SubClassWndEvent )
            {
                CloseHandle( g_SubClassWndEvent );
                g_SubClassWndEvent = NULL;
            }

        }
    }

    if ( g_OrigOutputWindowProc && g_hWndVboxOut )
    {
        SetWindowLongPtr(
            g_hWndVboxOut,
            GWLP_WNDPROC,
            (LONG_PTR)g_OrigOutputWindowProc
            );

        g_OrigOutputWindowProc = NULL;
    }

    DbgPrintf( "Exiting\n" );
}

VOID
OnDropFilesMsg(HDROP hDrop)
{

    UINT ClientId = GetWaitForInputFirstClientId();

    if ( ClientId == 0 )
    {
        DbgPrintf( "DragDrop: GuestServer is not connected" );
        return;
    }

    POINT pt = {0,0};

    GetCursorPos( &pt );
    ScreenToClient( g_hWndVboxOut, &pt );

    //FIXME: remove it
    CallWindowProc( g_OrigOutputWindowProc, g_hWndVboxOut, WM_MOUSEMOVE, 0, MAKELONG(pt.x, pt.y) );

    UINT uDrop = DragDrop_CreateHandle( hDrop );
   
    UINT uErr = HostCall_DragDropFiles(
        ClientId,
        uDrop
        );

    DbgPrintf( "WM_DROPFILES: Notify Guest Server: %u\n", uErr );

}

void
VM_EnableDragDrop( HWND hWndOut )
{
    g_hWndVboxOut = hWndOut;

    DbgPrintf( "VirtBoxOutputWnd = 0x%p\n", g_hWndVboxOut );
    DragAcceptFiles( g_hWndVboxOut, TRUE );

    g_OrigOutputWindowProc = (WNDPROC)SetWindowLongPtr(
        g_hWndVboxOut,
        GWLP_WNDPROC,
        (LONG_PTR)OutputWindowProcHook
        );

    DbgPrintf( 
        "SetWindowLongPtr(GWLP_WNDPROC) return %p (new=%p,%p)\n", 
        g_OrigOutputWindowProc, 
        (PVOID)GetWindowLongPtr( g_hWndVboxOut, GWLP_WNDPROC ),
        OutputWindowProcHook
        );

}

UINT
DragDrop_CreateHandle( HDROP hDrop )
{
    vector<wstring> FilesList;

    UINT Index = 0;
    WCHAR wszFileName[ MAX_PATH ] = L"";

    while ( DragQueryFileW( hDrop, Index, wszFileName, ARRAYSIZE(wszFileName) ) != 0 )
    {
        FilesList.push_back( wszFileName );
        Index++;
    }

    if ( FilesList.size() == 0 )
    {
        DbgPrintf( "DragDrop: FileList is zero length\n" );
        return (UINT)-1;
    }

    UINT RetHandle = (UINT)InterlockedIncrement( &g_DroppedFilesListLastHandle );

    g_DroppedFilesListLock.Lock();

    if ( g_DroppedFilesList.size() > 10 )
    {
        DbgPrintf( "Truncate g_DroppedFilesList\n" );
        g_DroppedFilesList.erase( g_DroppedFilesList.begin() );
    }

    g_DroppedFilesList[ RetHandle ] = FilesList;

    g_DroppedFilesListLock.Unlock();

    return RetHandle;
}

BOOL
GetDroppedFilesListByHandle(
    UINT DropHandle,
    vector<wstring> *pFileList
    )
{
    g_DroppedFilesListLock.Lock();

    map<UINT, vector<wstring> >::iterator iter = g_DroppedFilesList.find( DropHandle );

    if ( iter == g_DroppedFilesList.end() )
    {
        g_DroppedFilesListLock.Unlock();
        DbgPrintf( "GetDroppedFilesListByHandle: Cannot validate handle: %u\n", DropHandle );
        return FALSE;
    }

    *pFileList = iter->second;
    g_DroppedFilesListLock.Unlock();

    return TRUE;

}

BOOL
CALLBACK
DragDropGetCopiedFileListHandler(
    UINT DropHandle,
    vector<wstring> *pClipboardFileList
    )
{
    if ( DropHandle != 0 )
    {
        return GetDroppedFilesListByHandle( DropHandle, pClipboardFileList );
    }

    return ClipboardGetCopiedFileList( pClipboardFileList );
}


