/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "GuestServer.h"
#include "Globals.h"

#include "NewClipService.h"
#include "DbgPrintf.h"
#include "GuestServiceConnect.h"
#include "vboxhdrs.h"
#include "GuestCalls.h"
#include "PackedArgs.h"
#include "GServerMsgs.h"
#include "CopyFilesHandler.h"

#include "WinTools.h"
#include "Resource.h"
#include "DllMain.h"
#include "LangTools.h"
#include "DragDropGuest.h"
#include "ChkMAlloc.h"

#include <windows.h>

#define WM_NOTIFY_ICON (WM_APP+1)

#define MN_STATUS 1
#define MN_EXIT   2
#define MN_RESTARTASADMIN 3
#define MN_DRAGDROPFOLDER_CLEAN 4

VOID
AddNotifyIcon( HWND hwnd );

void
OnRestartAsAdmin();

void
OnDragDropFolderClean();

void
RunGuestServerProceedRequestsLoop();

VOID
InsertRestartAsAdminMenuItem(
    HMENU hMenu
    );

VOID
ShowContextMenu(
    HWND hWnd
    );

VOID
ShowStatusMessageBox();

void
DisconnectFromHost();

VOID
StopGServerWaitReceiveLoop();

VOID
__stdcall
StopGuestServer();

string
GetVerChkMsg(
    VERCHK_STATUS VerChkCode,
    UINT VerChkNativeErrCode
    );

void
RunGuestServerInternal();

BOOL
Is64bitOS();

LRESULT
CALLBACK
MonitorClipboardWndProc(
    HWND hwnd, 
    UINT msg, 
    WPARAM wParam, 
    LPARAM lParam
    );


DWORD
WINAPI
MonitorClipboardThread(
    LPVOID lpParameter
    );


VOID
RunMonitorClipboardThread();

VOID
OnClipboardContentWasChanged();

UINT
GuestServer_Process(
    UINT ConnectionId,
    CPackedArgs *pInputArgs,
    CPackedArgs *pOutputArgs
    );


volatile uint32_t g_ConnectionId = 0;
HWND g_NextClipbrdWnd = NULL;

VERCHK_STATUS g_VerChkCode = VERCHK_GEN_FAILURE;
UINT g_VerChkNativeErrCode = VINF_SUCCESS;
volatile HANDLE g_hMonitorClipboardThread = NULL;
BOOL g_StoppingServer = FALSE;
volatile BOOL g_ServerRunning = FALSE;
HWND g_MonitorClipboardWnd = NULL;
UINT g_NetDdeAgentDyingMsg = 0;
UINT g_TaskRecreatedMessage = 0;


VOID
RunMonitorClipboardThread()
{
    DWORD dwTid = 0;

    g_hMonitorClipboardThread = CreateThread(
        NULL,
        0,
        (LPTHREAD_START_ROUTINE)MonitorClipboardThread,
        NULL,
        0,
        &dwTid
        );

    if ( g_hMonitorClipboardThread == NULL )
    {
        DbgPrintf( "Cannot create MonitorClipboardThread: %u\n", GetLastError() );
    }
}


DWORD
WINAPI
MonitorClipboardThread(
    LPVOID lpParameter
    )
{
    HANDLE hCapturedHandle  = NULL;
    WNDCLASSEX wcex;
    HINSTANCE hInstance = GetModuleHandle(NULL);

    RtlZeroMemory( &wcex, sizeof(wcex) );

    wcex.cbSize        = sizeof(WNDCLASSEX);
    wcex.style         = 0;
    wcex.lpfnWndProc   = MonitorClipboardWndProc;
    wcex.hInstance     = hInstance;
    wcex.hIcon         = NULL;
    wcex.hCursor       = NULL;
    wcex.hbrBackground = NULL;
    wcex.lpszMenuName  = NULL;
    wcex.lpszClassName = GSERV_MONCLIPBOARD_CLASSA;

    if( !RegisterClassEx( &wcex ) )
    {
        DbgPrintf( "RegisterClassEx failed: %u\n", GetLastError() );

        hCapturedHandle = g_hMonitorClipboardThread;
        g_hMonitorClipboardThread = NULL;
        if ( hCapturedHandle )
            CloseHandle( hCapturedHandle );
        
        return 0;
    }

    HWND hWnd = CreateWindowEx(
        0,
        wcex.lpszClassName,
        "",
        WS_OVERLAPPEDWINDOW,
        0, 0, 0, 0,
        NULL,
        NULL, 
        hInstance,
        NULL
        );

    if ( NULL == hWnd )
    {
        DbgPrintf( "CreateWindowEx failed: %u\n", GetLastError() );

        hCapturedHandle = g_hMonitorClipboardThread;
        g_hMonitorClipboardThread = NULL;
        if ( hCapturedHandle )
            CloseHandle( hCapturedHandle );


        return 0;
    }

    ShowWindow( hWnd, SW_HIDE );

    g_NextClipbrdWnd = SetClipboardViewer( hWnd );

    MSG Msg;

    while( GetMessage( &Msg, NULL, 0, 0) )
    {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }

    hCapturedHandle = g_hMonitorClipboardThread;
    g_hMonitorClipboardThread = NULL;
    if ( hCapturedHandle )
        CloseHandle( hCapturedHandle );

    return 0;
}


LRESULT
CALLBACK
MonitorClipboardWndProc(
    HWND hwnd, 
    UINT msg, 
    WPARAM wParam, 
    LPARAM lParam
    )
{

    if ( msg == g_NetDdeAgentDyingMsg )
    {
        //WinNT4 shutdown support

        DbgPrintf( "NetDdeAgentDying message is riched\n" );
    
        DWORD dwOSVer = GetVersion();
        DWORD dwMajorVersion = (DWORD)(LOBYTE(LOWORD(dwOSVer)));

        if ( 4 == dwMajorVersion )
        {
            //Destroy Window if Current OS is WinNT4

            DestroyWindow( hwnd );
            return DefWindowProc(hwnd, msg, wParam, lParam );
        }
    }
	
	if ( msg == g_TaskRecreatedMessage && g_TaskRecreatedMessage )
	{
		AddNotifyIcon( hwnd );
	}

    switch( msg )
    {
        case WM_CREATE:
        {

            g_NetDdeAgentDyingMsg = RegisterWindowMessage( "NetDdeAgentDying" ); //0xc04b
			g_TaskRecreatedMessage = RegisterWindowMessage( "TaskbarCreated" );

            DbgPrintf( "NetDdeAgentDying = 0x%x\n",  g_NetDdeAgentDyingMsg );

            g_MonitorClipboardWnd = hwnd;
			AddNotifyIcon( hwnd );

            return DefWindowProc(hwnd, msg, wParam, lParam);
            break;
        }

        case WM_NOTIFY_ICON:
        {
            switch ( lParam )
            {
                case WM_RBUTTONDOWN:
                {
                    ShowContextMenu( hwnd );
                    break;
                }

                case WM_LBUTTONDOWN:
                {
                    ShowStatusMessageBox();
                    break;
                }
            }
            break;
        }

        case WM_COMMAND:
        {
            UINT idCmd = LOWORD( wParam );

            switch ( idCmd )
            {
                case MN_EXIT:
                {
                    DestroyWindow( hwnd );
                    break;
                }

                case MN_STATUS:
                {
					Dbg_GuestCall_DumpInfo();
                    ShowStatusMessageBox();
                    break;
                }

                case MN_RESTARTASADMIN:
                {
                    OnRestartAsAdmin();
                    break;
                }
				
				case MN_DRAGDROPFOLDER_CLEAN:
				{
					OnDragDropFolderClean();
					break;
				}
            }

            break;
        }

        case WM_CHANGECBCHAIN:
        {
            HWND hwndToRemove = (HWND)wParam;
            HWND hwndNext = (HWND)lParam;

            if ( g_NextClipbrdWnd == hwndToRemove )
            {
                g_NextClipbrdWnd = hwndNext;
            }
            else if ( g_NextClipbrdWnd != NULL )
            {
                SendMessage( g_NextClipbrdWnd, msg, wParam, lParam );
            }

            break;
        }

        case WM_DRAWCLIPBOARD:
        {

            OnClipboardContentWasChanged();

            if ( g_NextClipbrdWnd != NULL )
            {
                SendMessage( g_NextClipbrdWnd, msg, wParam, lParam );
            }

            break;
        }

        case WM_DESTROY:
        {
            g_MonitorClipboardWnd = NULL;

            SetClipboardViewer( g_NextClipbrdWnd );
            NotifyIcon_Delete( hwnd );

            StopGServerWaitReceiveLoop();

            //PostQuitMessage(0);
			ExitProcess(0); //force exit process
            break;
        }

        case WM_CLOSE:
        {
            DestroyWindow( hwnd );
            break;
        }

        case WM_QUERYENDSESSION:
        {
            //DestroyWindow( hwnd );
            return TRUE;
            break;
        }

        default:
        {
            return DefWindowProc(hwnd, msg, wParam, lParam);
        }
    }
    return 0;
}

VOID
OnClipboardContentWasChanged()
{
    if ( !OpenClipboard(NULL) )
        return;

    BOOL bAvail = IsClipboardFormatAvailable( CF_HDROP );
    CloseClipboard();

    if ( !bAvail )
    {
        DbgPrintf( "There are no files in the clipboard\n" );
        return;
    }

    DbgPrintf( "There are files in the clipboard!\n" );

    if ( 0 == g_ConnectionId )
    {
        DbgPrintf( "There is no connection to the host\n" );
        return;
    }

    UINT result = GuestCall_HSM_NotifyGuestClipboardHasFiles( g_ConnectionId, (UINT)GetVersion(), (UINT)Is64bitOS() );

    if ( !RT_SUCCESS( result ) )
    {
        DbgPrintf( "NotifyGuestClipboardHasFiles failed: %d\n", result );
    }
}

void
RunGuestServerProceedRequestsLoop()
{
    g_ServerRunning = TRUE;
    g_StoppingServer = FALSE;
   
    int code = VMInit();
    
    if ( RT_FAILURE( code ) )
    {
        DbgPrintf( "VMInit failed: %d\n", code );
        g_ServerRunning = FALSE;
        return;
    }

    
    code = CONNECT_FUNCTION( (PUINT)&g_ConnectionId );

    if ( !RT_SUCCESS( code ) )
    {
        DbgPrintf( "CONNECT_FUNCTION failed: %d\n", code );
        g_ServerRunning = FALSE;
        return;
    }

    
    UINT InputBufferSize = GSERV_MAX_INPUT_BUFFER_SIZE;

    PVOID pInputBuffer = MALLOC( InputBufferSize );

    if ( NULL == pInputBuffer )
    {
        DbgPrintf( "Allocation of memory for Input Buffer failed\n" );
        DisconnectFromHost();
        return;
    }

    while ( TRUE )
    {
        UINT PacketId = 0;
        UINT ReadInputBufferSize = 0;

        UINT result = GuestCall_HSM_WaitForInputForGuestServer(
            g_ConnectionId,
            pInputBuffer,
            InputBufferSize,
            &ReadInputBufferSize,
            &PacketId
            );

        if ( g_StoppingServer )
            break;

        if ( result != ERROR_SUCCESS )
        {
            DbgPrintf( "GuestCall_HSM_WaitForInputForGuestServer failed: %d\n", result );
            break;
        }

        CPackedArgs InputArgs;
        CPackedArgs OutputArgs;

        if ( !InputArgs.Unpack( pInputBuffer, ReadInputBufferSize ) )
        {
            DbgPrintf( "Cannot unpack args\n" );
            break;
        }

        UINT ErrorCode = GuestServer_Process(
            g_ConnectionId,
            &InputArgs,
            &OutputArgs
            );

        if ( g_StoppingServer )
            break;

        PVOID pPackedBuff = NULL;
        ULONG PackedBuffSize = 0;

        if ( !OutputArgs.Pack( &pPackedBuff, &PackedBuffSize ) )
        {
            DbgPrintf( "Cannot pack args\n" );
            break;
        }

        result = GuestCall_HSM_SendOutputFromGuestServer(
            g_ConnectionId,
            &ErrorCode,
            pPackedBuff,
            PackedBuffSize,
            PacketId
            );

        CPackedArgs::FreeMem( pPackedBuff );

        if ( g_StoppingServer )
            break;


    } // while


    FREE( pInputBuffer );


    DisconnectFromHost();
    g_ServerRunning = FALSE;

}

void
RunGuestServerInternal()
{
    DbgPrintf( "RunGuestServer called\n" );

    if ( g_ServerRunning )
    {
        DbgPrintf( "Server is already running\n" );
        return;
    }
	
	//Clean up...
	
	wstring wDirToClean = GetDragDropTmpMainFolder();
	
	if ( wDirToClean.length() > 0 )
	{
		RemoveDirectoryRecursiveW( wDirToClean.c_str(), APP_NAMEW, TRUE );
	}

    UINT HostHookVersion = 0;

    g_VerChkCode = GuestCall_CheckHostHookVersion( &HostHookVersion, &g_VerChkNativeErrCode );

    if ( g_VerChkCode != VERCHK_SUCCESS )
    {
        DbgPrintf( "GuestCall_CheckHostHookVersion failed: %u, native err code: %d\n", (UINT)g_VerChkCode, g_VerChkNativeErrCode );
    }

    RunMonitorClipboardThread();
    RunGuestServerProceedRequestsLoop();
}

UINT
GuestServer_Process(
    UINT ConnectionId,
    CPackedArgs *pInputArgs,
    CPackedArgs *pOutputArgs
    )
{

    if ( pInputArgs->GetCount() == 0 )
    {
        DbgPrintf( "ERROR: Input args count = 0\n" );
        return ERROR_INVALID_PARAMETER;
    }

    UINT uMessage = 0;

    if ( !pInputArgs->GetUInt32Arg( 0, &uMessage ) )
    {
        DbgPrintf( "ERROR: Cannot get MessageId\n" );
        return ERROR_INVALID_PARAMETER;
    }

    switch ( uMessage )
    {
        case GSM_CFH_BEGIN:
        {
            if ( pInputArgs->GetCount() != GSM_CFH_BEGIN_IN_CPARAMS )
            {
                DbgPrintf( "ERROR: GSM_CFH_BEGIN: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32DropHandle = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32DropHandle ) )
            {
                DbgPrintf( "ERROR: Cannot get u32DropHandle\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32Handle = 0;

            UINT u32ErrorCode = CFH_Begin( ConnectionId, u32DropHandle, &u32Handle );

            pOutputArgs->CreateFromArgs2(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode,
                CPackedArgs::ARGT_UINT32, &u32Handle
                );

            break;
        }

        case GSM_CFH_END:
        {
            if ( pInputArgs->GetCount() != GSM_CFH_END_IN_CPARAMS )
            {
                DbgPrintf( "ERROR: GSM_CFH_END: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32Handle = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32Handle ) )
            {
                DbgPrintf( "ERROR: Cannot get u32Handle\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32ErrorCode = CFH_End( ConnectionId, u32Handle );

            pOutputArgs->CreateFromArgs1(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode
                );

            break;
        }

        case GSM_CFH_GET_BUILDING_LIST_STATUS:
        {

            if ( pInputArgs->GetCount() != GSM_CFH_GET_BUILDING_LIST_STATUS_IN_CPARAMS )
            {
                DbgPrintf( "ERROR: GSM_CFH_GET_BUILDING_LIST_STATUS: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32Handle = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32Handle ) )
            {
                DbgPrintf( "ERROR: Cannot get u32Handle\n" );
                return ERROR_INVALID_PARAMETER;
            }

            ULONG StatusBufferSize = sizeof( COPY_FILES_BUILDING_LIST_STATUS );

            PCOPY_FILES_BUILDING_LIST_STATUS pStatus = (PCOPY_FILES_BUILDING_LIST_STATUS)CPackedArgs::AllocateMem(
                StatusBufferSize
                );

            if ( NULL == pStatus )
            {
                return ERROR_NOT_ENOUGH_MEMORY;
            }

            RtlZeroMemory( pStatus, StatusBufferSize );
            pStatus->StructSize = sizeof(COPY_FILES_BUILDING_LIST_STATUS); //FIXED

            UINT u32ErrorCode = CFH_GetBuildingListStatus(
                ConnectionId,
                u32Handle,
                pStatus
                );

            pOutputArgs->CreateFromArgs2(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode,
                CPackedArgs::ARGT_DYNAMIC_PTR, &PTR_STRUCT(StatusBufferSize, pStatus)
                );

            break;
        }

        case GSM_CFH_GETCURRENTFILENAME:
        {

            DbgPrintf( "GSM_CFH_GETCURRENTFILENAME\n" );

            if ( pInputArgs->GetCount() != GSM_CFH_GETCURRENTFILENAME_IN_CPARAMS )
            {
                DbgPrintf( "ERROR: GSM_CFH_GETCURRENTFILENAME: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32Handle = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32Handle ) )
            {
                DbgPrintf( "ERROR: Cannot get u32Handle\n" );
                return ERROR_INVALID_PARAMETER;
            }

            ULONG GetCurrFileNameBuffSize = sizeof(COPY_FILES_GET_CURRENT_FILENAME);

            PCOPY_FILES_GET_CURRENT_FILENAME pGetCurrFileName = (PCOPY_FILES_GET_CURRENT_FILENAME)CPackedArgs::AllocateMem(
                GetCurrFileNameBuffSize
                );

            if ( NULL == pGetCurrFileName )
            {
                return ERROR_NOT_ENOUGH_MEMORY;
            }

            RtlZeroMemory( pGetCurrFileName, GetCurrFileNameBuffSize );
            pGetCurrFileName->StructSize = sizeof(COPY_FILES_GET_CURRENT_FILENAME); //FIXED

            UINT u32ErrorCode = CFH_GetCurrentFileName(
                ConnectionId,
                u32Handle,
                pGetCurrFileName
                );

            DbgPrintf( "CFH_GetCurrentFileName return: %u\n", u32ErrorCode );

            pOutputArgs->CreateFromArgs2(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode,
                CPackedArgs::ARGT_DYNAMIC_PTR, &PTR_STRUCT(GetCurrFileNameBuffSize, pGetCurrFileName)
                );


            break;
        }

        case GSM_CFH_GETCURRENTSTREAM:
        {
            if ( pInputArgs->GetCount() != GSM_CFH_GETCURRENTSTREAM_IN_CPARAMS  )
            {
                DbgPrintf( "ERROR: GSM_CFH_GETCURRENTSTREAM: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
                
            }

            UINT u32Handle = 0;
            UINT u32BufferSize = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32Handle ) ||
                 !pInputArgs->GetUInt32Arg( 2, &u32BufferSize ) ||
                 u32BufferSize > GSERV_MAX_OUTPUT_STREAM_SIZE
                )
            {
                DbgPrintf( "ERROR: GSM_CFH_GETCURRENTSTREAM: invalid params\n" );
                return ERROR_INVALID_PARAMETER;
            }

            PVOID pStream = (PCOPY_FILES_BUILDING_LIST_STATUS)CPackedArgs::AllocateMem(
                u32BufferSize
                );

            if ( NULL == pStream )
            {
                return ERROR_NOT_ENOUGH_MEMORY;
            }

            ULONG BytesCopied = 0;
            UINT64 u64StreamOffset = 0;

            UINT u32ErrorCode = CFH_GetCurrentStream(
                ConnectionId,
                u32Handle,
                pStream,
                u32BufferSize,
                (PUINT)&BytesCopied,
                &u64StreamOffset
                );

            pOutputArgs->CreateFromArgs3(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode,
                CPackedArgs::ARGT_DYNAMIC_PTR, &PTR_STRUCT(BytesCopied, pStream),
                CPackedArgs::ARGT_UINT64, &u64StreamOffset
                );


            break;
        }

        case GSM_CFH_MOVETONEXTITEM:
        {
            if ( pInputArgs->GetCount() != GSM_CFH_MOVETONEXTITEM_IN_CPARAMS  )
            {
                DbgPrintf( "ERROR: GSM_CFH_MOVETONEXTITEM: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
                
            }

            UINT   u32Handle = 0;
            UINT64 u64ItemId = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &u32Handle ) ||
                 !pInputArgs->GetUInt64Arg( 2, &u64ItemId )
                )
            {
                DbgPrintf( "ERROR: GSM_CFH_MOVETONEXTITEM: invalid params\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32ErrorCode = CFH_MoveToNextItem(
                ConnectionId,
                u32Handle,
                u64ItemId
                );

            pOutputArgs->CreateFromArgs1(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode
                );


            break;
        }

        case GSM_DROP_FILES:
        {
            if ( pInputArgs->GetCount() != GSM_DROP_FILES_IN_CPARAMS )
            {
                DbgPrintf( "ERROR: GSM_CFH_BEGIN: bad params count\n" );
                return ERROR_INVALID_PARAMETER;
            }

            UINT u32ErrorCode = ERROR_SUCCESS;

            UINT   uDropHandle = 0;

            if ( !pInputArgs->GetUInt32Arg( 1, &uDropHandle ) )
            {
                DbgPrintf( "ERROR: GSM_DROP_FILES: invalid params\n" );
                return ERROR_INVALID_PARAMETER;
            }

            GSM_AsyncOnDropFiles( uDropHandle );

            pOutputArgs->CreateFromArgs1(
                CPackedArgs::ARGT_UINT32, &u32ErrorCode
                );

            break;
        }


        default:
        {
            DbgPrintf( "ERROR: Unknown MessageId: %u\n", uMessage );
            return ERROR_INVALID_PARAMETER;
            break;
        }
    }

    return ERROR_SUCCESS;
}


BOOL
Is64bitOS()
{
    typedef
    UINT 
    (WINAPI *GetSystemWow64DirectoryAProc)(
        LPTSTR lpBuffer,
        UINT uSize
        );

    GetSystemWow64DirectoryAProc pGetSystemWow64DirectoryA = (GetSystemWow64DirectoryAProc)GetProcAddress(
            GetModuleHandle( "kernel32.dll" ),
            "GetSystemWow64DirectoryA"
            );
    
    if ( pGetSystemWow64DirectoryA == NULL )
        return FALSE;

    CHAR szBuff[MAX_PATH] = "";

    UINT Ret = pGetSystemWow64DirectoryA(
        szBuff,
        MAX_PATH
        );
    
    if ( Ret == 0 && GetLastError() == ERROR_CALL_NOT_IMPLEMENTED )
        return FALSE;

    return TRUE;
}

string
GetVerChkMsg(
    VERCHK_STATUS VerChkCode,
    UINT VerChkNativeErrCode
    )
{

    string result;

    switch ( VerChkCode )
    {
        case VERCHK_SUCCESS:
        {
            result = LangGetStringA( IDS_VERCHK_SUCCESS );
            break;
        }

        case VERCHK_VER_MISMATCH_CURRENT_IS_OLDER:
        {
            result = LangGetStringA( IDS_VER_MISMATCH_CURRENT_IS_OLDER );
            break;
        }

        case VERCHK_VER_MISMATCH_CURRENT_IS_NEWER:
        {
            result = LangGetStringA( IDS_VER_MISMATCH_CURRENT_IS_NEWER );
            break;
        }

        case VERCHK_TOOLS_ARE_NOT_INSTALLED:
        {
            result = LangGetStringA( IDS_VER_TOOLS_ARE_NOT_INSTALLED );
            break;
        }

        case VERCHK_HOSTHOOK_IS_NOT_INSTALLED:
        {
            result = LangGetStringA( IDS_VERCHK_HOSTHOOK_IS_NOT_INSTALLED );
            break;
        }


        default:
        {
            result = LangFormatA( IDS_VERCHK_GEN_ERROR_FMT, (UINT)VerChkCode, (UINT)VerChkNativeErrCode );
            break;
        }

    }


    return result;
}


EXTERN_C
void
__stdcall
RunGuestServer()
{

    if ( OpenAppMutexAndCheckIsAppInstanceExist( GSERV_APP_MUTEXA ) )
    {
        DbgPrintf( "Guest server already launched\n" );
        return;
    }

    RunGuestServerInternal();

    //give ability to tray icon to display error status

    if ( g_hMonitorClipboardThread )
    {
        WaitForSingleObject( g_hMonitorClipboardThread, INFINITE );

        if ( g_hMonitorClipboardThread )
        {
            CloseHandle( g_hMonitorClipboardThread );
        }
    }
}

void
DisconnectFromHost()
{
    uint32_t CapturedConnectionId = InterlockedExchange( (LONG *)&g_ConnectionId, 0 );

    if ( 0 ==  CapturedConnectionId )
        return;

    DISCONNECT_FUNCTION( CapturedConnectionId  );

}


VOID
StopGServerWaitReceiveLoop()
{
    if ( g_StoppingServer )
        return;

    g_StoppingServer = TRUE;
    DisconnectFromHost();
    
}

VOID
__stdcall
StopGuestServer()
{
    StopGServerWaitReceiveLoop();

    if ( g_MonitorClipboardWnd )
    {
        SendMessage( g_MonitorClipboardWnd, WM_CLOSE, 0, 0 );
    }
}


VOID
ShowContextMenu(
    HWND hWnd
    )
{
    HMENU hMenu = CreatePopupMenu();

	if( NULL == hMenu )
        return;


    InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION,
        MN_STATUS,
        LangGetStringA( IDS_STATUS ).c_str()
        );

    InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION | MF_SEPARATOR,
        0,
        ""
        );

    InsertRestartAsAdminMenuItem(
        hMenu
        );
		
    InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION,
        MN_DRAGDROPFOLDER_CLEAN,
        LangGetStringA( IDS_DRAGDROPFOLDER_CLEAN ).c_str()
        );
		
	InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION | MF_SEPARATOR,
        0,
        ""
        );
		

    InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION,
        MN_EXIT,
        LangGetStringA( IDS_EXIT ).c_str()
        );


    POINT pt;
	GetCursorPos( &pt );

    SetForegroundWindow( hWnd );

    TrackPopupMenu(
        hMenu,
        TPM_BOTTOMALIGN,
		pt.x,
        pt.y,
        0,
        hWnd,
        NULL
        );

    DestroyMenu( hMenu );
}


VOID
ShowStatusMessageBox()
{
    UINT HostHookVersion = 0;
    UINT VerChkNativeErrCode = 0;

    VERCHK_STATUS VerChkCode = GuestCall_CheckHostHookVersion( &HostHookVersion, &VerChkNativeErrCode );

    MessageBox( NULL, GetVerChkMsg( VerChkCode, VerChkNativeErrCode ).c_str(), APP_NAMEA, MB_SETFOREGROUND | MB_APPLMODAL | MB_ICONINFORMATION  );

}

VOID
InsertRestartAsAdminMenuItem(
    HMENU hMenu
    )
{
    DWORD dwVersion = GetVersion();
    DWORD dwMajorVersion = (DWORD)( LOBYTE(LOWORD(dwVersion)) );

    //Show item for Vista+

    if ( dwMajorVersion < 6 )
        return;

    string GServAsAdminFullPath = GetCurrentProcessDirA() + string( "\\" ) + string( GSERV_ASADMIN_EXENAMEA );

    if ( !IsFileExistA( GServAsAdminFullPath.c_str() ) )
        return;

    if ( 0 == _stricmp( GetCurrentProcessFullPathA().c_str(), GServAsAdminFullPath.c_str() ) )
        return;

    BOOL AdmEnabled = FALSE;

    if ( IsAdminGroupEnabled( &AdmEnabled ) && AdmEnabled )
        return;

    InsertMenu(
        hMenu,
        -1,
        MF_BYPOSITION,
        MN_RESTARTASADMIN,
        LangGetStringA( IDS_RESTARTASADMIN ).c_str()
        );
    

}

void
OnRestartAsAdmin()
{
    DbgPrintf( "OnRestartAsAdmin\n" );

    string GServAsAdminFullPath = GetCurrentProcessDirA() + string( "\\" ) + string( GSERV_ASADMIN_EXENAMEA );

    StopGServerWaitReceiveLoop();

    ULONG TryCount = 20;

    while ( TryCount-- > 0 && g_ServerRunning )
    {
        Sleep( 60 );
    }

    if ( g_ServerRunning )
    {
        DbgPrintf( "OnRestartAsAdmin: Server was not stopped\n" );
    }

    CloseAppMutex();

/*    PROCESS_INFORMATION pi;
    STARTUPINFO si;

    RtlZeroMemory( &si, sizeof(si) );

    si.cb = sizeof( si );

    BOOL bRet = CreateProcess(
        NULL,
        (LPSTR)GServAsAdminFullPath.c_str(),
        NULL,
        NULL,
        FALSE,
        0,
        NULL,
        NULL,
        &si,
        &pi
        ); */

    DWORD dwRet = (DWORD)ShellExecute(
        NULL,
        "open",
        GServAsAdminFullPath.c_str(),
        NULL,
        NULL,
        SW_SHOWNORMAL
        );

    if ( dwRet > 32 )
    {
        SendMessage( g_MonitorClipboardWnd, WM_CLOSE, 0, 0 );
        ExitProcess( ERROR_SUCCESS );
        return;
    }

    //Resume

    DbgPrintf( "ShellExecute failed: %u\n", dwRet );

    OpenAppMutexAndCheckIsAppInstanceExist( GSERV_APP_MUTEXA );
    RunGuestServerProceedRequestsLoop();
}

VOID
AddNotifyIcon( HWND hwnd )
{
	HICON hIcon = NULL;

	if ( g_VerChkCode == VERCHK_SUCCESS )
		hIcon = LoadIcon( g_DllHinst, MAKEINTRESOURCE( IDI_GSERV_ENABLED ) );
	else
		hIcon = LoadIcon( g_DllHinst, MAKEINTRESOURCE( IDI_GSERV_DISABLED ) );

	NotifyIcon_Add( hwnd, WM_NOTIFY_ICON, hIcon, APP_NAMEA );
}

void
OnDragDropFolderClean()
{
	wstring wDirToClean = GetDragDropTmpMainFolder();
	
	if ( wDirToClean.length() == 0 )
		return;
	
    DWORD dwRet  = MessageBoxW( 
		NULL, 
		LangFormatW( IDS_CONFIRM_DRAGDROPFOLDER_CLEAN_FMTW, wDirToClean.c_str() ).c_str(), 
		APP_NAMEW, 
		MB_SETFOREGROUND | MB_APPLMODAL | MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2
		);
		
	if ( dwRet != IDYES )
		return;


	RemoveDirectoryRecursiveW( wDirToClean.c_str(), APP_NAMEW, FALSE );
	
}

