/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "PathFromPoint.h"
#include <shlwapi.h>
#include <shlobj.h>
#include "DbgPrintf.h"

//==================================


wstring
PathFromPoint(
    POINT pt
	);

wstring
COM_Explorer_PathFromPoint( POINT pt );


//wstring
//Desktop_PathFromPoint( POINT pt );

HWND
GetTopParentWindow( HWND hWnd );



//==================================

HWND
GetTopParentWindow( HWND hWnd )
{
	DWORD dwWndPID = 0;
	GetWindowThreadProcessId( hWnd, &dwWndPID );

	if ( 0 == dwWndPID )
		return hWnd;

	HWND hWndParent = hWnd;
	HWND hWndTop = hWnd;

	while ( (hWndParent = GetParent(hWndParent)) != NULL )
	{
		DWORD dwPidParent = 0;

		GetWindowThreadProcessId( hWndParent, &dwPidParent );

		if ( dwPidParent != dwWndPID )
			break;

		hWndTop = hWndParent;
	}

	return hWndTop;

}

wstring
COM_Explorer_PathFromPoint( POINT pt )
{
	wstring result = L"";

	HWND hWndFromPoint = WindowFromPoint( pt );

	if ( NULL == hWndFromPoint )
    {
        DbgPrintf( "COM_Explorer_PathFromPoint: hWndFromPoint == NULL \n" );
		return L"";
    }

	HWND hWndTop = GetTopParentWindow( hWndFromPoint );

	//printf( "hWndTop = 0x%p\n", hWndTop );

    IShellWindows *pShellWindows = NULL;

    HRESULT hr = CoCreateInstance(
		CLSID_ShellWindows, 
		NULL, 
		CLSCTX_LOCAL_SERVER, 
		IID_IShellWindows, 
		(PVOID *)(&pShellWindows)
		);


	if ( !SUCCEEDED(hr) )
	{
		DbgPrintf( "CoCreateInstance failed: 0x%x\n", hr );
		return L"";
	}


	IDispatch  *pDispatch = NULL;
	BOOL bFound = FALSE;

	VARIANT vItemNumber;
	vItemNumber.vt = VT_I4;

	for ( 
	  vItemNumber.lVal = 0; 
	  !bFound && pShellWindows->Item( vItemNumber, &pDispatch ) == S_OK;
	   vItemNumber.lVal++
	  ) 
	{
	IWebBrowserApp *pWebBrowserApp = NULL;

	if (SUCCEEDED( pDispatch->QueryInterface(IID_IWebBrowserApp, (void**)&pWebBrowserApp )) ) 
	{
		HWND WebBrowserhWnd = NULL;

		if ( SUCCEEDED(pWebBrowserApp->get_HWND( (LONG_PTR*)&WebBrowserhWnd )  ) )
		{
			//printf( "HWND = %p\n", WebBrowserhWnd );
		}

		IServiceProvider *pServiceProvider = NULL;

		if ( WebBrowserhWnd == hWndTop && SUCCEEDED(pWebBrowserApp->QueryInterface(IID_IServiceProvider, (void**)&pServiceProvider)) ) 
		{
			IShellBrowser *pShellBrowser = NULL;

			if (SUCCEEDED(pServiceProvider->QueryService(SID_STopLevelBrowser,
							  IID_IShellBrowser, (void**)&pShellBrowser ))) 
			{
				IShellView *pShellView = NULL;

			if ( SUCCEEDED(pShellBrowser->QueryActiveShellView( &pShellView ) ) ) 
			{
				IFolderView *pFolderView = NULL;

				if (SUCCEEDED( pShellView->QueryInterface(IID_IFolderView,
											   (void**)&pFolderView))) 
				{
					IPersistFolder2 *pPersistFolder2 = NULL;


					if (SUCCEEDED(pFolderView->GetFolder(IID_IPersistFolder2,
											(void**)&pPersistFolder2))) 
					{
						LPITEMIDLIST pidlFolder;

						if (SUCCEEDED( pPersistFolder2->GetCurFolder(&pidlFolder)) ) 
						{
							WCHAR szPath[ MAX_PATH ] = L"";

							if ( !SHGetPathFromIDListW(pidlFolder, szPath) ) 
							{
								//printf( "<NO PATH>\n" );
							}
							else
							{
								//printf( "PATH: %ws\n", szPath );
								result = szPath;
								bFound = TRUE;
							}

							CoTaskMemFree( pidlFolder );
						}					

						pPersistFolder2->Release();
					}

					pFolderView->Release();
				}

				pShellView->Release();
			}

				pShellBrowser->Release();
			}



			pServiceProvider->Release();
		}


		pWebBrowserApp->Release();
	}


	pDispatch->Release();
	}


	pShellWindows->Release();


	return result;
}

wstring
PathFromPoint(
    POINT pt
	)
{
	wstring result;

    HRESULT hr = CoInitializeEx(
		NULL,
		COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE
		);

    if ( !SUCCEEDED(hr) )
    {
        DbgPrintf( "PathFromPoint: CoInitializeEx failed: 0x%x\n", hr );
    }

    DbgPrintf( "PathFromPoint: CoInitializeEx: Success \n" );
	
	result = COM_Explorer_PathFromPoint( pt );

    CoUninitialize();

	//if ( result == L"" )
	//	result = Desktop_PathFromPoint( pt );


	
	return result;
}

/*wstring
Desktop_PathFromPoint( POINT pt )
{
	wstring result;

	HWND hWndFromPoint = WindowFromPoint( pt );

	if ( NULL == hWndFromPoint )
		return L"";

	printf( "hWndFromPoint: %p DesktopWindow: %p\n", hWndFromPoint, GetDesktopWindow() );

	if ( hWndFromPoint == GetDesktopWindow() )
	{
		result = GetDesktopFolder();
	}

	return result;
}*/

