
/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


#include "VBoxUtil.h"
#include "DbgPrintf.h"
#include "ComUtil.h"

	
HRESULT
Find_MachineDisp(
	IN DWORD dwMachinePid,
	OUT IDispatch **ppMachineDisp
	);
	
HRESULT
Variant_ConvertToUINT(
	IN VARIANT *pInput,
	OUT UINT *pOutput
	);	
	

IDispatch *g_pCurrentMachineObj = NULL;

VOID
VBoxUtil_Init()
{
	static BOOL s_WasInit = FALSE;
	
	if ( s_WasInit )
		return;
			
	HRESULT hr = Find_MachineDisp(GetCurrentProcessId(), &g_pCurrentMachineObj );
	
	if ( !SUCCEEDED(hr) )
	{
		g_pCurrentMachineObj = NULL;
	}
		
	s_WasInit = TRUE;	
}


VOID
VBoxUtil_Clear()
{
	if ( g_pCurrentMachineObj )
	{
		g_pCurrentMachineObj->Release();
		g_pCurrentMachineObj = NULL;
	}
}

HRESULT
Find_MachineDisp(
	IN DWORD dwMachinePid,
	OUT IDispatch **ppMachineDisp
	)
{
	IDispatch *pdispVirtualBox = NULL;

	HRESULT hr = CreateInstanceWithProgID(
		L"VirtualBox.VirtualBox",
		CLSCTX_LOCAL_SERVER,
		&pdispVirtualBox
		);

	if ( !SUCCEEDED( hr ) )
	{
		return hr;
	}

	VARIANT varMachinesArr;
	VariantInit( &varMachinesArr );

	hr = Disp_PropGet(
		&varMachinesArr,
		pdispVirtualBox,
		L"machines"
		);

	if ( SUCCEEDED( hr ) )
	{
		if ( (VT_ARRAY | VT_DISPATCH) != varMachinesArr.vt )
		{
			hr = E_INVALIDARG;
		}

		if ( SUCCEEDED( hr ) )
		{

			SAFEARRAY *saMachines = (SAFEARRAY *)varMachinesArr.parray;

			IDispatch **DispMachines = NULL;

			hr = SafeArrayAccessData ( saMachines, (void **) &DispMachines );

			if ( SUCCEEDED( hr ) )
			{
				BOOL bFound = FALSE;				

				for ( ULONG u = 0; u < saMachines->rgsabound[0].cElements; u++ )
				{
					//sessionPid SessionPID

					VARIANT varMachinePID;
					VariantInit( &varMachinePID );

					UINT uCurrentMachinePid = 0;

					hr = Disp_PropGet(
						&varMachinePID,
						DispMachines[u],
						L"sessionPid"
						);

					if ( SUCCEEDED( hr ) )
					{
						hr = Variant_ConvertToUINT( &varMachinePID, &uCurrentMachinePid );

						if ( !SUCCEEDED( hr ) )
						{
							VariantClear( &varMachinePID );
						}
					}

					if ( !SUCCEEDED( hr ) )
						break;

					bFound = (uCurrentMachinePid == dwMachinePid);
					VariantClear( &varMachinePID );

					if ( bFound )
					{
						*ppMachineDisp = DispMachines[u];
						(*ppMachineDisp)->AddRef();
						break;
					}
				}

				if ( SUCCEEDED( hr ) && !bFound )
					hr = E_FAIL;

				SafeArrayUnaccessData( saMachines );
			}

		}

		VariantClear( &varMachinesArr );		
	}

	pdispVirtualBox->Release();
	return hr;

}

HRESULT
Variant_ConvertToUINT(
	IN VARIANT *pInput,
	OUT UINT *pOutput
	)
{
	//TODO: VariantChangeType

	HRESULT hr = S_OK;

	switch ( pInput->vt )
	{
		case VT_UI4:
		{
			*pOutput = V_UI4( pInput );
			break;
		}

		case VT_UINT:
		{
			*pOutput = V_UINT( pInput );
			break;
		}

		case VT_I4:
		{
			*pOutput = V_I4( pInput );
			break;
		}

		case VT_INT:
		{
			*pOutput = V_INT( pInput );
			break;
		}

		case VT_BOOL:
		{
			*pOutput = V_BOOL( pInput );
			break;
		}

		case VT_UI2:
		{
			*pOutput = V_UI2( pInput );
			break;
		}

		case VT_I2:
		{
			*pOutput = V_I2( pInput );
			break;
		}

		case VT_UI1:
		{
			*pOutput = V_UI1( pInput );
			break;
		}

		case VT_I1:
		{
			*pOutput = V_I1( pInput );
			break;
		}

		default:
		{
			hr = E_INVALIDARG;
			break;
		}
	}
	
	return hr;
}


DWORD
WINAPI
VBoxUtil_GetClipboardModeWorkerThread( PVOID pParam )
{
	UINT* ClipMode = (UINT*)pParam;
	
	if ( g_pCurrentMachineObj == NULL )
		return FALSE;
	
	VARIANT varClipboardMode;
	VariantInit( &varClipboardMode );

	HRESULT rc = Disp_PropGet(
		&varClipboardMode,
		g_pCurrentMachineObj,
		L"clipboardMode"
		);
		
	if ( SUCCEEDED( rc ) )
	{
		rc = Variant_ConvertToUINT( &varClipboardMode, ClipMode );
		VariantClear( &varClipboardMode );
	}	
	
	return SUCCEEDED( rc );
	
}


BOOL
VBoxUtil_GetClipboardMode( OUT UINT *pClipboardMode )
{	
	if ( g_pCurrentMachineObj == NULL )
	{
		return FALSE;
	}
	
	UINT ClipMode = CLIPBOARD_MODE_BIDIRECTIONAL;
	VARIANT varClipboardMode;
	VariantInit( &varClipboardMode );

	HRESULT rc = Disp_PropGet(
		&varClipboardMode,
		g_pCurrentMachineObj,
		L"clipboardMode"
		);
		
	if ( SUCCEEDED( rc ) )
	{
		rc = Variant_ConvertToUINT( &varClipboardMode, &ClipMode );
		VariantClear( &varClipboardMode );
	}
	
	if ( rc == RPC_E_CANTCALLOUT_ININPUTSYNCCALL ) 
	{
		//FIXED: Paste from Guest to Host
		//RPC_E_CANTCALLOUT_ININPUTSYNCCALL
		//http://support.microsoft.com/kb/198996/
		
		DWORD dwTid = 0;
		
		HANDLE hThread = CreateThread(
			NULL,
			0,
			VBoxUtil_GetClipboardModeWorkerThread,
			&ClipMode,
			0,
			&dwTid
			);
			
		if ( hThread == NULL )
			return FALSE;
			
		WaitForSingleObject( hThread, INFINITE );
		
		DWORD dwExitCodeThread = 0;
		
		GetExitCodeThread( hThread, &dwExitCodeThread );
		
		if ( dwExitCodeThread )
		{
			*pClipboardMode = (DWORD)ClipMode;
		}
		
		CloseHandle( hThread );		
		return dwExitCodeThread;
		
	}
	
	if ( SUCCEEDED( rc ) )
	{
		if ( ClipMode <= CLIPBOARD_MODE_MAX )
		{
			*pClipboardMode = (DWORD)ClipMode;
			return TRUE;
		}			
	}
	else
	{
		DbgPrintf( "VBoxUtil_GetClipboardMode get_ClipboardMode failed: %x\n", rc );
	}


	return FALSE;	
}

BOOL
VBoxUtil_GetMachineName( OUT wstring *pMachineName )
{
	if ( g_pCurrentMachineObj == NULL )
	{
		return FALSE;
	}
	
	VARIANT varMachineName;
	VariantInit( &varMachineName );

	HRESULT hr = Disp_PropGet(
		&varMachineName,
		g_pCurrentMachineObj,
		L"name"
		);

	if ( !SUCCEEDED( hr ) )
		return FALSE;
	
	if ( varMachineName.vt == VT_BSTR )
	{
		*pMachineName = V_BSTR( &varMachineName );
		VariantClear( &varMachineName );
		return TRUE;
	}

	VariantClear( &varMachineName );
	return FALSE;
}

