/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


#include <windows.h>
#include <wincrypt.h>
#include <wintrust.h>
#include <SoftPub.h>
#include "DbgPrintf.h"
#include "Globals.h"


typedef
CRYPT_PROVIDER_CERT* (WINAPI * TWTHelperGetProvCertFromChain)(
  CRYPT_PROVIDER_SGNR *pSgnr,
  DWORD idxCert
  );

typedef
CRYPT_PROVIDER_DATA* (WINAPI *TWTHelperProvDataFromStateData)(
	HANDLE hStateData
	);
	
typedef	
CRYPT_PROVIDER_SGNR* (WINAPI *TWTHelperGetProvSignerFromChain)(
  CRYPT_PROVIDER_DATA *pProvData,
  DWORD idxSigner,
  BOOL fCounterSigner,
  DWORD idxCounterSigner
 );	
 
typedef
DWORD (WINAPI *TCertGetNameStringW)(
  PCCERT_CONTEXT pCertContext,
  DWORD dwType,
  DWORD dwFlags,
  void *pvTypePara,
  LPWSTR pszNameString,
  DWORD cchNameString
  ); 
 
//==================== 
 
//These functions are present in WinXP+ only

CRYPT_PROVIDER_CERT*
WINAPI
Dyn_WTHelperGetProvCertFromChain(
  CRYPT_PROVIDER_SGNR *pSgnr,
  DWORD idxCert
  );
  
CRYPT_PROVIDER_DATA* 
WINAPI 
Dyn_WTHelperProvDataFromStateData(
	HANDLE hStateData
	);
  
CRYPT_PROVIDER_SGNR* 
WINAPI 
Dyn_WTHelperGetProvSignerFromChain(
  CRYPT_PROVIDER_DATA *pProvData,
  DWORD idxSigner,
  BOOL fCounterSigner,
  DWORD idxCounterSigner
  );

DWORD
WINAPI
Dyn_CertGetNameStringW(
  PCCERT_CONTEXT pCertContext,
  DWORD dwType,
  DWORD dwFlags,
  void *pvTypePara,
  LPWSTR pszNameString,
  DWORD cchNameString
  );
  
  
//====================   

EXTERN_C
BOOL 
WINAPI 
DllMain(
    HINSTANCE hinstDLL,
    DWORD fdwReason,
    LPVOID lpvReserved
    )
{
    if ( fdwReason == DLL_PROCESS_ATTACH )
    {
        
    }

    return TRUE;
}

BOOL
AnsiToUnicode(
    LPCSTR lpszSource,
    LPWSTR lpDestBuff,
    SIZE_T BuffSizeInWords
    )
{
    if ( 0 == BuffSizeInWords )
        return FALSE;

    int ret = MultiByteToWideChar(
        GetACP(),
        0,
        lpszSource,
        -1,
        lpDestBuff,
        BuffSizeInWords
        );

    lpDestBuff[ BuffSizeInWords - 1 ] = 0;

    return ret != 0;

}


EXTERN_C
__declspec(dllexport)
BOOL
CALLBACK
CheckSigner(
    LPCSTR lpszFileName,
    LPCSTR lpSignerName
    )
{
    
    WINTRUST_DATA WintrustData;
    GUID ActionID = WINTRUST_ACTION_GENERIC_VERIFY_V2;
    WCHAR wszSigner[ 200 ] = L"";
    WCHAR wszFileName[ MAX_PATH ] = L"";
    BOOL bResult = FALSE;

    AnsiToUnicode( lpszFileName, wszFileName, ARRAYSIZE(wszFileName) );
    AnsiToUnicode( lpSignerName, wszSigner, ARRAYSIZE(wszSigner) );

    WINTRUST_FILE_INFO FileInfo ;

    RtlZeroMemory( &FileInfo, sizeof(FileInfo) );
    FileInfo.cbStruct = sizeof(WINTRUST_FILE_INFO);

    FileInfo.hFile = NULL;
    FileInfo.pcwszFilePath = wszFileName;
    FileInfo.pgKnownSubject = NULL;

    RtlZeroMemory( &WintrustData, sizeof(WintrustData) );

    WintrustData.cbStruct = sizeof(WintrustData);

    WintrustData.fdwRevocationChecks = WTD_REVOKE_NONE;
    WintrustData.dwUnionChoice       = WTD_CHOICE_FILE;
    WintrustData.dwStateAction       = WTD_STATEACTION_VERIFY;
    WintrustData.dwUIChoice = WTD_UI_NONE;

    WintrustData.pFile = &FileInfo;

    //WTD_HASH_ONLY_FLAG
    WintrustData.dwProvFlags = WTD_CACHE_ONLY_URL_RETRIEVAL | WTD_REVOCATION_CHECK_NONE;

    HRESULT hr = WinVerifyTrust((HWND)INVALID_HANDLE_VALUE, &ActionID, &WintrustData);

    DbgPrintf( "WinVerifyTrust return %x\n", hr );

    if ( ( hr != S_OK ) && ( hr != CERT_E_EXPIRED ) )
        return FALSE;

    CRYPT_PROVIDER_DATA const *psProvData     = NULL;
    CRYPT_PROVIDER_SGNR       *psProvSigner   = NULL;
    CRYPT_PROVIDER_CERT       *psProvCert     = NULL;

    const CRYPT_PROVIDER_DATA * pProvData = Dyn_WTHelperProvDataFromStateData ( WintrustData.hWVTStateData );

    if ( NULL != pProvData )
    {
        CRYPT_PROVIDER_SGNR * pProvSigner = Dyn_WTHelperGetProvSignerFromChain ((PCRYPT_PROVIDER_DATA)pProvData, 0, FALSE, 0);

        if ( pProvSigner != NULL )
        {
            CRYPT_PROVIDER_CERT * pProvCert = Dyn_WTHelperGetProvCertFromChain (pProvSigner, 0);

            if ( pProvCert != NULL )
            {                
                DWORD dwTypePara = CERT_X500_NAME_STR;

                DWORD dwBuffSize = Dyn_CertGetNameStringW (
                    pProvCert->pCert, 
                    CERT_NAME_SIMPLE_DISPLAY_TYPE, 
                    0, 
                    &dwTypePara, 
                    NULL, 
                    0);

                if ( dwBuffSize > 0 )
                {
                    PWCHAR pSignerName = (PWCHAR)LocalAlloc (0, dwBuffSize * sizeof (WCHAR ));

                    if ( pSignerName )
                    {

                        dwBuffSize = Dyn_CertGetNameStringW(
                            pProvCert->pCert,
                            CERT_NAME_SIMPLE_DISPLAY_TYPE,
                            0,
                            &dwTypePara,
                            pSignerName,
                            dwBuffSize
                            );
                        
                        if ( dwBuffSize > 0 && 0 == _wcsicmp( pSignerName, wszSigner ) )
                        {
                            bResult = TRUE;
                        }

                        LocalFree ( pSignerName );
                    }
                }
            }
        }
    
    } //
    
    WintrustData.dwStateAction = WTD_STATEACTION_CLOSE;
    WinVerifyTrust ((HWND)INVALID_HANDLE_VALUE, &ActionID, &WintrustData );


    return bResult;
}

VOID
WINAPI
GS_HelperTerminateAppInstance()
{
	HWND hAppWnd = FindWindow(
		GSERV_MONCLIPBOARD_CLASSA,
		NULL
		);

	if ( hAppWnd == NULL )
		return;

	DWORD dwAppPid = 0;
	GetWindowThreadProcessId( hAppWnd, &dwAppPid );

	SendMessage(
		hAppWnd,
		WM_CLOSE,
		0,
		0
		);

	DWORD dwEndTick = GetTickCount() + 4000;

	while ( GetTickCount() < dwEndTick )
	{		
		Sleep( 60 );

		DWORD dwPid = 0;

		GetWindowThreadProcessId( hAppWnd, &dwPid );

		if ( dwPid != dwAppPid )
			break;
	}
}


CRYPT_PROVIDER_CERT*
WINAPI
Dyn_WTHelperGetProvCertFromChain(
  CRYPT_PROVIDER_SGNR *pSgnr,
  DWORD idxCert
  )
{
	HINSTANCE hWinTrustDll = GetModuleHandle( "Wintrust.dll" );
	
	if ( hWinTrustDll == NULL )
		hWinTrustDll = LoadLibrary(  "Wintrust.dll" );
		
	if ( hWinTrustDll == NULL )
		return NULL;
		
	TWTHelperGetProvCertFromChain pfnOrig = (TWTHelperGetProvCertFromChain)GetProcAddress( hWinTrustDll, "WTHelperGetProvCertFromChain" );
	
	if ( pfnOrig == NULL )
		return NULL;
		
	return pfnOrig( pSgnr, idxCert );
	
	
}  


CRYPT_PROVIDER_DATA* 
WINAPI 
Dyn_WTHelperProvDataFromStateData(
	HANDLE hStateData
	)
{
	HINSTANCE hWinTrustDll = GetModuleHandle( "Wintrust.dll" );
	
	if ( hWinTrustDll == NULL )
		hWinTrustDll = LoadLibrary(  "Wintrust.dll" );
		
	if ( hWinTrustDll == NULL )
		return NULL;
		
	TWTHelperProvDataFromStateData pfnOrig = (TWTHelperProvDataFromStateData)GetProcAddress( hWinTrustDll, "WTHelperProvDataFromStateData" );
	
	if ( pfnOrig == NULL )
		return NULL;
		
	return pfnOrig( hStateData );
	

}
	

CRYPT_PROVIDER_SGNR* 
WINAPI 
Dyn_WTHelperGetProvSignerFromChain(
  CRYPT_PROVIDER_DATA *pProvData,
  DWORD idxSigner,
  BOOL fCounterSigner,
  DWORD idxCounterSigner
 )
{
	HINSTANCE hWinTrustDll = GetModuleHandle( "Wintrust.dll" );
	
	if ( hWinTrustDll == NULL )
		hWinTrustDll = LoadLibrary(  "Wintrust.dll" );
		
	if ( hWinTrustDll == NULL )
		return NULL;
		
	TWTHelperGetProvSignerFromChain pfnOrig = (TWTHelperGetProvSignerFromChain)GetProcAddress( hWinTrustDll, "WTHelperGetProvSignerFromChain" );
	
	if ( pfnOrig == NULL )
		return NULL;
		
	return pfnOrig(
		pProvData,
		idxSigner,
		fCounterSigner,
		idxCounterSigner
		);

}


DWORD
WINAPI
Dyn_CertGetNameStringW(
  PCCERT_CONTEXT pCertContext,
  DWORD dwType,
  DWORD dwFlags,
  void *pvTypePara,
  LPWSTR pszNameString,
  DWORD cchNameString
  )
{
	HINSTANCE hCryptDll = GetModuleHandle( "Crypt32.dll" );
	
	if ( hCryptDll == NULL )
		hCryptDll = LoadLibrary(  "Crypt32.dll" );
		
	if ( hCryptDll == NULL )
		return NULL;
		
	TCertGetNameStringW pfnOrig = (TCertGetNameStringW)GetProcAddress( hCryptDll, "CertGetNameStringW" );
	
	if ( pfnOrig == NULL )
		return NULL;

	return pfnOrig(
		pCertContext,
		dwType,
		dwFlags,
		pvTypePara,
		pszNameString,
		cchNameString
		);

}  
  
 
 

