/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <windows.h>
#include "ConfirmDlg.h"
#include "resource.h"
#include <string>
#include "LangTools.h"
#include "SecureFuncs.h"
#include "ChkMAlloc.h"
#include <strsafe.h>

using namespace std;

typedef struct _DLG_FILE_IO_ERROR_PARAMS
{
	wstring FileName;
	wstring Msg1;
	ULONG ErrorCode;

} DLG_FILE_IO_ERROR_PARAMS, *PDLG_FILE_IO_ERROR_PARAMS;

typedef struct _DLG_OVERWRITE_CONFIRM_PARAMS
{
	wstring SrcFileName;
	FILETIME SrcLastWriteTime;
	ULONG64 SrcFileSize;

	wstring DestFileName;
	FILETIME DestLastWriteTime;
	ULONG64 DestFileSize;

}DLG_OVERWRITE_CONFIRM_PARAMS, *PDLG_OVERWRITE_CONFIRM_PARAMS;

//==============================================

wstring
ErrorCodeToStringW(
    ULONG ErrorCode
	);

wstring
FileSizeToStr(
    ULONG64 FileSize
	);

//==============================================

wstring
FileSizeToStr(
    ULONG64 FileSize
	)
{
	wstring result;
	WCHAR wszStr[200] = L"";

	if ( FileSize >= 1024*1024*1024 )
	{
		StringCchPrintfW(
			wszStr,
			ARRAYSIZE(wszStr),
			L"%.2f Gb",
			double(FileSize) / double( 1024*1024*1024 )
			);

	}
	else if ( FileSize >= 1024*1024 )
	{
		StringCchPrintfW(
			wszStr,
			ARRAYSIZE(wszStr),
			L"%.2f Mb",
			double(FileSize) / double( 1024*1024 )
			);

	}
	else if ( FileSize >= 1024 )
	{
		StringCchPrintfW(
			wszStr,
			ARRAYSIZE(wszStr),
			L"%.2f Kb",
			double(FileSize) / double( 1024 )
			);

	}
	else
	{
		StringCchPrintfW(
			wszStr,
			ARRAYSIZE(wszStr),
			LangGetStringW(IDS_BYTES_FMT).c_str(),
			FileSize
			);

	}

	result = wszStr;

	return result;

}

wstring
FormatFileDateTimeAndSize(
	FILETIME *pFileTime,
	ULONG64 FileSize
	)
{
	wstring result;

	SYSTEMTIME LocalTime;

	FileTimeToSystemTime( pFileTime, &LocalTime );

	SystemTimeToTzSpecificLocalTime(
		NULL,
		&LocalTime,
		&LocalTime
		);

	WCHAR szTime[200] = L"<no time>";

	GetTimeFormatW(
      LOCALE_USER_DEFAULT,
      0,
      &LocalTime,
      NULL, //L"hh':'mm':'ss", //it needs because there is no leading zero when it is NULL,
      szTime,
      ARRAYSIZE(szTime)
      );


	WCHAR szDate[200] = L"<no date>";

	GetDateFormatW(
      LOCALE_USER_DEFAULT,
      DATE_LONGDATE,
      &LocalTime,
      NULL,
      szDate,
      ARRAYSIZE(szDate)
	  );

	result += FileSizeToStr(FileSize);
	result += L",";

	result += szDate;
	result += L",";

	result += szTime;



	return result;
}


INT_PTR 
CALLBACK 
OverwriteDialogProc(
  HWND hwndDlg,
  UINT uMsg,
  WPARAM wParam,
  LPARAM lParam
	)
{
	if ( uMsg == WM_INITDIALOG )
	{
		if ( lParam == 0 )
			return FALSE;

		PDLG_OVERWRITE_CONFIRM_PARAMS pParams = (PDLG_OVERWRITE_CONFIRM_PARAMS)lParam;

		SetWindowTextW(
			hwndDlg,
			LangGetStringW( IDS_DLG_REPL ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			GRP_OVERWRITEFILE_DEST,
			LangGetStringW( IDS_OVERWRITEFILE_DEST ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			GRP_OVERWRITEFILE_SRC,
			LangGetStringW( IDS_OVERWRITEFILE_SRC ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			IDOK,
			LangGetStringW( IDS_YES ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			ID_YES_TO_ALL,
			LangGetStringW( IDS_YES_TO_ALL ).c_str()
			);


		SetDlgItemTextW(
			hwndDlg,
			ID_NO,
			LangGetStringW( IDS_NO ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			ID_NO_TO_ALL,
			LangGetStringW( IDS_NO_TO_ALL ).c_str()
			);

		SetDlgItemTextW(
			hwndDlg,
			IDCANCEL,
			LangGetStringW( IDS_CANCEL ).c_str()
			);


		SetDlgItemTextW( hwndDlg, IDC_DESTFILENAME, pParams->DestFileName.c_str() );
		SetDlgItemTextW( hwndDlg, IDC_SRCFILENAME, pParams->SrcFileName.c_str() );

		SetDlgItemTextW( 
				hwndDlg, IDC_DESTFDATA, 
				FormatFileDateTimeAndSize( &pParams->DestLastWriteTime, pParams->DestFileSize ).c_str() 
				);


		SetDlgItemTextW( 
				hwndDlg, IDC_SRCFDATA, 
				FormatFileDateTimeAndSize( &pParams->SrcLastWriteTime, pParams->SrcFileSize ).c_str()
				);


		delete pParams;
		return TRUE;
	}

	if ( uMsg == WM_COMMAND )
	{
		switch ( LOWORD(wParam) )
		{
		case IDCANCEL:
			{
				EndDialog( hwndDlg, DLG_CANCEL );
				break;
			}

		case IDOK:
			{
				EndDialog( hwndDlg, DLG_YES );
				break;
			}

		case ID_NO:
			{
				EndDialog( hwndDlg, DLG_NO );
				break;
			}
		case ID_YES_TO_ALL:
			{
				EndDialog( hwndDlg, DLG_YES_TO_ALL );
				break;
			}

		case ID_NO_TO_ALL:
			{
				EndDialog( hwndDlg, DLG_NO_TO_ALL );
				break;
			}

		}
	}

	return FALSE;
}


int
DlgOverwriteConfirm(
	HINSTANCE hInst,
	LPCWSTR lpszSrcFileName,
	FILETIME *pSrcLastWriteTime,
	ULONG64 SrcFileSize,
	LPCWSTR lpszDestFileName,
	FILETIME *pDestLastWriteTime,
	ULONG64 DestFileSize
	)
{
	PDLG_OVERWRITE_CONFIRM_PARAMS DlgParams = new DLG_OVERWRITE_CONFIRM_PARAMS;

	if ( DlgParams == NULL )
		return DLG_CANCEL;

	DlgParams->SrcFileName = lpszSrcFileName;
	DlgParams->SrcLastWriteTime = *pSrcLastWriteTime;
	DlgParams->SrcFileSize = SrcFileSize;

	DlgParams->DestFileName = lpszDestFileName;
	DlgParams->DestLastWriteTime = *pDestLastWriteTime;
	DlgParams->DestFileSize = DestFileSize;

	int ret = (int)DialogBoxParam(
		hInst,
		MAKEINTRESOURCE(IDD_DLG_REPL),
		NULL,
		OverwriteDialogProc,
		(LPARAM)DlgParams //lParam
		);

	if ( ( (ret == 0 ) || ( ret == -1 ) ) && ERROR_SUCCESS != GetLastError() )
	{
		delete DlgParams;
		return DLG_CANCEL;
	}

	return ret;

}

int
DlgOverwriteConfirmCached(
	HINSTANCE hInst,
	LPCWSTR lpszSrcFileName,
	FILETIME *pSrcLastWriteTime,
	ULONG64 SrcFileSize,
	LPCWSTR lpszDestFileName,
	FILETIME *pDestLastWriteTime,
	ULONG64 DestFileSize,
	int *pCachedValue
	)
{
	if ( *pCachedValue != 0 )
		return *pCachedValue;

	int ret = DlgOverwriteConfirm(
		hInst,
		lpszSrcFileName,
		pSrcLastWriteTime,
		SrcFileSize,
		lpszDestFileName,
		pDestLastWriteTime,
		DestFileSize
		);

	if ( ret == DLG_YES_TO_ALL || ret == DLG_NO_TO_ALL )
	{
		*pCachedValue = ret;
	}

	return ret;
}

wstring
ErrorCodeToStringW(
    ULONG ErrorCode
	)
{
	LPWSTR strMsg = NULL;
	wstring result;

	int ret = FormatMessageW(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        ErrorCode,
        0,
        (LPWSTR)&strMsg,
        0,
        NULL);

	if ( ret == 0 )
	{
		WCHAR wszBuff[200] = L"";

		StringCchPrintfW(
			wszBuff,
			ARRAYSIZE(wszBuff),
			LangGetStringW( IDS_ERROR_CODE_FMT ).c_str(),
			ErrorCode
			);

		return wszBuff;
	}

	result = strMsg;
	LocalFree( strMsg );
	return result;

}



INT_PTR 
CALLBACK 
FileIoErrorDialogProc(
  HWND hwndDlg,
  UINT uMsg,
  WPARAM wParam,
  LPARAM lParam
  )
{
	if ( uMsg == WM_INITDIALOG )
	{
		if ( lParam == 0 )
			return FALSE;

		PDLG_FILE_IO_ERROR_PARAMS pParams = (PDLG_FILE_IO_ERROR_PARAMS)lParam;

		SetWindowTextW( hwndDlg, LangGetStringW( IDS_ERROR ).c_str() );
		LangSetDlgItemTextW( hwndDlg, ID_SKIP, IDS_SKIP );
		LangSetDlgItemTextW( hwndDlg, ID_SKIP_ALL, IDS_SKIP_ALL );
		LangSetDlgItemTextW( hwndDlg, ID_RETRY, IDS_RETRY );
		LangSetDlgItemTextW( hwndDlg, IDCANCEL, IDS_CANCEL );

		SetDlgItemTextW( hwndDlg, GRP_MSG1, pParams->Msg1.c_str() );
		SetDlgItemTextW( hwndDlg, IDC_FILENAME, pParams->FileName.c_str() );
		SetDlgItemTextW( hwndDlg, IDC_MSG2, ErrorCodeToStringW( pParams->ErrorCode ).c_str() );

		delete pParams;

		return TRUE;
	}

	if ( uMsg == WM_COMMAND )
	{
		switch ( LOWORD(wParam) )
		{
		case IDCANCEL:
			{
				EndDialog( hwndDlg, DLG_CANCEL );
				break;
			}

		case ID_SKIP:
			{
				EndDialog( hwndDlg, DLG_SKIP );
				break;
			}

		case ID_SKIP_ALL:
			{
				EndDialog( hwndDlg, DLG_SKIP_ALL );
				break;
			}

		case ID_RETRY:
			{
				EndDialog( hwndDlg, DLG_RETRY );
				break;
			}
		}
	}

	return FALSE;
}

int
DlgFileIoError(
	HINSTANCE hInst,
	LPCWSTR lpszFileName,
	LPCWSTR lpMsg1,
	ULONG IoErrorCode
	)
{
	int ret = DLG_CANCEL;

	PDLG_FILE_IO_ERROR_PARAMS pParams = new DLG_FILE_IO_ERROR_PARAMS;

	if ( NULL == pParams )
		return DLG_CANCEL;

	pParams->ErrorCode = IoErrorCode;
	pParams->FileName = lpszFileName;
	pParams->Msg1 = lpMsg1;

	ret = (int)DialogBoxParam(
		hInst,
		MAKEINTRESOURCE(IDD_FILEIO_ERROR),
		NULL,
		FileIoErrorDialogProc,
		(LPARAM)pParams
		);

	if ( ( (ret == 0 ) || ( ret == -1 ) ) && ERROR_SUCCESS != GetLastError() )
	{
		delete pParams;
		return DLG_CANCEL;
	}


	return ret;
}


int
DlgFileIoErrorCached(
	HINSTANCE hInst,
	LPCWSTR lpszFileName,
	LPCWSTR lpMsg1,
	ULONG IoErrorCode,
	//should be init with zero
	int *pCachedValue

	)
{
	if ( *pCachedValue != 0 )
		return *pCachedValue;

	int ret = DlgFileIoError(
		hInst,
		lpszFileName,
		lpMsg1,
		IoErrorCode
		);

	if ( ret == DLG_SKIP_ALL )
	{
		*pCachedValue = ret;
	}

	return ret;

}

