/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "DbgPrintf.h"
#include <stdio.h>
#include <stdarg.h>
#include <windows.h>
#include "SecureFuncs.h"
#include "ChkMAlloc.h"
#include <strsafe.h>

VOID
PutDebugMessageToFile( CHAR *pString );

HANDLE g_hDebugFile = INVALID_HANDLE_VALUE;


#ifndef USE_NT_DBG_PRINTF

void
DbgPrintfReal(
    const char *Format,
    ...        
    )
{
    va_list vaList;
    CHAR szBuf[512] = "";
 
    va_start( vaList, Format );

    StringCchVPrintf( szBuf, ARRAYSIZE(szBuf), Format, vaList );
    szBuf[ARRAYSIZE(szBuf) - 1 ] = 0;
        
    OutputDebugStringA( szBuf );
	PutDebugMessageToFile( szBuf );
}

#endif


void
NtDbgPrintfReal(
    const char *Format,
    ...        
    )
{
    typedef
    int
    (__cdecl * T_vsnprintf)(
        char *buffer,
        size_t count,
        const char *format,
        va_list argptr 
        );

    HINSTANCE hNtdll = GetModuleHandle( "ntdll.dll" );

    if ( NULL == hNtdll )
    {
        OutputDebugStringA( "NtDbgPrintfReal: Cannot get instance of ntdll.dll\n" );
        return;
    }

    T_vsnprintf _vsnprintfProc = (T_vsnprintf)GetProcAddress( hNtdll, "_vsnprintf" );

    if ( NULL == _vsnprintfProc )
    {
        OutputDebugStringA( "NtDbgPrintfReal: Cannot determine address of _vsnprintf\n" );
        return;
    }


    va_list vaList;
    CHAR szBuf[512];

    szBuf[0] = 0;
 
    va_start( vaList, Format );

    _vsnprintfProc( szBuf, ARRAYSIZE(szBuf), Format, vaList );
    szBuf[ARRAYSIZE(szBuf) - 1 ] = 0;
        
    OutputDebugStringA( szBuf );
	PutDebugMessageToFile( szBuf );

}

VOID
PutDebugMessageToFile( CHAR *pString )
{

	CHAR szDebugFileName[MAX_PATH] = "";
	
	if ( g_hDebugFile == INVALID_HANDLE_VALUE )
	{
		GetModuleFileName( GetModuleHandle(NULL), szDebugFileName, ARRAYSIZE(szDebugFileName) );
		StringCchCatA( szDebugFileName, ARRAYSIZE(szDebugFileName), ".debug.log" );
		
		g_hDebugFile = CreateFile(
			szDebugFileName,
			GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL
			);
	}
	
	if ( g_hDebugFile == INVALID_HANDLE_VALUE )
		return;
		
	SYSTEMTIME stTime;
	CHAR szTime[ 100 ] = "";
	
	GetLocalTime( &stTime );
	
	LONG cchReturn = GetDateFormat(
		LOCALE_USER_DEFAULT,
		DATE_SHORTDATE,
		&stTime,
		NULL,
		szTime,
		ARRAYSIZE(szTime)
		);
	
	if ( cchReturn  != 0 )
		szTime[ cchReturn - 1 ] = ' ';
		
	GetTimeFormat(
		LOCALE_USER_DEFAULT,
		0, //TIME_NOSECONDS,
		&stTime,
		NULL,
		&szTime[cchReturn],
		ARRAYSIZE(szTime) - cchReturn
		);		
		
	StringCchCatA( szTime, ARRAYSIZE(szTime), " " );
		
	SetFilePointer( g_hDebugFile, 0, NULL, FILE_END );
	
	DWORD dwWriteln = 0;
	
	WriteFile(
		g_hDebugFile,
		szTime,
		strlen(szTime),
		&dwWriteln,
		NULL
		);	
	
	
	WriteFile(
		g_hDebugFile,
		pString,
		strlen(pString),
		&dwWriteln,
		NULL
		);	
}
