/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "FilesCopier.h"
#include "FileOperationsProgressDlgs.h"
#include "DbgPrintf.h"
#include "LangTools.h"
#include "ConfirmDlg.h"
#include "WinTools.h"
#include "ChkMAlloc.h"
//#include "DllMain.h" //g_DllHinst

extern
HINSTANCE 
CALLBACK 
GetModuleInstance();


typedef enum _FILE_OP
{
    FILE_OP_CONTINUE,
    FILE_OP_SKIP,
    FILE_OP_CANCEL_ALL

} FILE_OP;

ULONG
ProceeedCopyingFiles(
    CFileCopierHelper *pCopyingHelper,
    uint32_t ClientId,
    LPCWSTR lpszDestFolderName
    );

FILE_OP
ProceedWriteStream(
    CFileCopierHelper *pCopyingHelper,
    IN  UINT ClientId,
	IN  UINT Handle,
    IN  OUT HANDLE &FileHandle,
    IN  PVOID pFileBuffer,
    IN  ULONG FileBufferSize,
    COM_CFileCopyingProgressDlg *pProgressDlg,
    IN  wstring & CurrentFileName,
    IN OUT ULONG64 &ProcessedItemsSize,
    IN COPY_FILES_BUILDING_LIST_STATUS &StatusInfo,
    IN int *pDlgFileWriteCached,
    IN COPY_FILES_GET_CURRENT_FILENAME &CurrFileName,
    IN OUT int *pDlgFileReadFromHostCached
    );

FILE_OP
ProceedFileCreate(
    HANDLE *pFileHandle,
	LPCWSTR lpszDestFolderName,
    LPCWSTR lpszRelFilePath,
    COPY_FILES_GET_CURRENT_FILENAME *pCurrFileName,
    int *pDlgOverWriteCached,
    int *pDlgFileCreateCached
    )
{
    wstring CurrentFileName = GetFullPathFromRelative( lpszDestFolderName, lpszRelFilePath );
    BOOL bOverwrite = FALSE;
    DWORD dwGle = 0;

    while ( TRUE )
    {
    	BOOL bRet = CreateFileWithRelPathW(
    	    pFileHandle,
            bOverwrite,
            &CurrentFileName,
    		lpszDestFolderName,
    		lpszRelFilePath,
    		pCurrFileName->FileAttr
    		);

        dwGle = GetLastError();

        if ( !bRet &&
             ERROR_FILE_EXISTS == dwGle &&
             !( pCurrFileName->FileAttr & FILE_ATTRIBUTE_DIRECTORY )
           )
        {
            UINT64 FileSize = 0;
            DWORD FileAttr = 0;
            FILETIME FileLastWriteTime;

            if ( GetFileInfoW( CurrentFileName.c_str(), &FileSize, &FileAttr, &FileLastWriteTime  ) )
            {
                wstring ClipboardFilePath = LangGetStringW( IDS_CLIPBOARDPATH_PREFIX );
                ClipboardFilePath += pCurrFileName->wszRelFilePath;

                int DlgOverWrite = DlgOverwriteConfirmCached(
                    GetModuleInstance(),
                    ClipboardFilePath.c_str(),
                    &pCurrFileName->FileLastWriteTime,
                    pCurrFileName->FileSize,
                    CurrentFileName.c_str(),
                    &FileLastWriteTime,
                    FileSize,
                    pDlgOverWriteCached
                    );

                if ( ( DLG_YES == DlgOverWrite ) || ( DlgOverWrite == DLG_YES_TO_ALL ) )
                {
                    bOverwrite = TRUE;

                	bRet = CreateFileWithRelPathW(
                	    pFileHandle,
                        bOverwrite,
                        &CurrentFileName,
                		lpszDestFolderName,
                		lpszRelFilePath,
                		pCurrFileName->FileAttr
                		);

                    dwGle = GetLastError();
                }
                else if ( ( DLG_NO == DlgOverWrite ) || ( DlgOverWrite == DLG_NO_TO_ALL ) )
                {
                    return FILE_OP_SKIP;
                }
                else
                {
                    return FILE_OP_CANCEL_ALL;
                }
            } // GetFileInfoW
        }

        if ( !bRet )
        {
            int FileCreateFailDlgRet = DlgFileIoErrorCached(
                GetModuleInstance(),
                CurrentFileName.c_str(),
                LangGetStringW( IDS_CANNOT_CREATE_FILE ).c_str(),
                dwGle,
                pDlgFileCreateCached
                );

            if ( FileCreateFailDlgRet == DLG_SKIP || FileCreateFailDlgRet == DLG_SKIP_ALL )
            {
                return FILE_OP_SKIP;
            }
            else if ( FileCreateFailDlgRet == DLG_RETRY )
            {
                continue;
            }
            else
            {
                return FILE_OP_CANCEL_ALL;
            }

        }

        if ( bRet )
            return FILE_OP_CONTINUE;


    } // while

    return FILE_OP_CONTINUE;


}


ULONG
ProceeedCopyingFiles(
    CFileCopierHelper *pCopyingHelper,
    uint32_t ClientId,
    uint32_t DropHandle,
    LPCWSTR lpszDestFolderName
    )
{
	ULONG ReturnResult = 0;
    uint32_t Handle = 0;
    uint32_t result = pCopyingHelper->CFH_Begin( ClientId, DropHandle, &Handle );
    
    DbgPrintf( "CFH_Begin return: %u Handle: %u\n", result, (UINT)Handle );
    
    if ( result != ERROR_SUCCESS )
    {
        if ( result == ERROR_NO_DATA )
        {
            LangErrorMessageBox( IDS_NOFILES_IN_CLIPBOARD );
        }
        else if ( ERROR_INVALID_OPERATION == result )
        {
            LangErrorMessageBox( IDS_INIT_FAIL_INV_OP );
        }
		else if ( ERROR_RESOURCE_DISABLED == result )
		{
			LangErrorMessageBox( IDS_CLIP_COPY_IS_NOT_ALLOWED );
		}
        else
        {
            LangErrorMessageBoxFmt( IDS_CONNECT_TO_HOST_FAIL_FMT, result );
        }
        return 0;
    }
    
    COPY_FILES_BUILDING_LIST_STATUS StatusInfo;

    RtlZeroMemory( &StatusInfo, sizeof(StatusInfo) );
    StatusInfo.StructSize = sizeof( StatusInfo ); 

    COM_CPrepareFileCopyingDlg PrepareDlg;

    while ( pCopyingHelper->CFH_GetBuildingListStatus( ClientId, Handle, &StatusInfo ) == ERROR_IO_PENDING )
	{
		//printf( "Scanned %u Files and %u Folders \r", StatusInfo.FileCount, StatusInfo.FolderCount );

        PrepareDlg.ShowProgress(
            NULL,
            StatusInfo.FileCount, 
            StatusInfo.FolderCount,
            StatusInfo.AllStreamSize,
            lpszDestFolderName
            );

        if ( PrepareDlg.IsCancelled() )
        {
            pCopyingHelper->CFH_End( ClientId, Handle );
            return 0;
        }


		Sleep( 60 );
	}
    
    PrepareDlg.EndDialog();

    DbgPrintf( "%u Files and %u Folders were found\n", StatusInfo.FileCount, StatusInfo.FolderCount );

    //=======================
    

    DWORD dwErr = ERROR_SUCCESS;
    
    ULONG FileBufferSize = 1024*400;
    PVOID pFileBuffer = MALLOC( FileBufferSize );

    if ( NULL == pFileBuffer )
    {   
        DbgPrintf( "Memory error\n" );
        return 0;
    }

    COM_CFileCopyingProgressDlg FileCopyingProgressDlg;
    ULONG64 ProcessedItemsSize = 0;
    BOOL bDone = FALSE;
    int DlgOverWriteCached = 0;
    int DlgFileCreateFailCached = 0;
    int DlgFileWriteCached = 0;
    int DlgFileReadFromHostCached = 0;


    while ( !bDone  )
	{
		COPY_FILES_GET_CURRENT_FILENAME CurrFileName;

		CurrFileName.StructSize = sizeof( CurrFileName );
		CurrFileName.wszRelFilePath[0] = 0;

		dwErr = pCopyingHelper->CFH_GetCurrentFileName(
			ClientId,
			Handle,
			&CurrFileName
			);

		if ( dwErr == ERROR_NO_MORE_FILES )
			break;

		if ( dwErr != ERROR_SUCCESS )
		{
			DbgPrintf( "CFH_GetCurrentFileName failed: %u\n", dwErr );
            LangErrorMessageBoxFmt( IDS_FUNC_FAILED_FMT, L"CFH_GetCurrentFileName", dwErr );
			break;
		}

		HANDLE FileHandle = INVALID_HANDLE_VALUE;
        wstring CurrentFileName = GetFullPathFromRelative( lpszDestFolderName, CurrFileName.wszRelFilePath );

        FILE_OP FileCreateResult = ProceedFileCreate(
            &FileHandle,
            lpszDestFolderName,
            CurrFileName.wszRelFilePath,
            &CurrFileName,
            &DlgOverWriteCached,
            &DlgFileCreateFailCached
            );
			
		if ( FileCreateResult == FILE_OP_CONTINUE )
		{
			ReturnResult++;
		}

        if ( FileCreateResult == FILE_OP_CANCEL_ALL )
        {
            break;
        }

        if ( FileCreateResult == FILE_OP_SKIP )
        {
            ProcessedItemsSize += CurrFileName.FileSize;

            dwErr = pCopyingHelper->CFH_MoveToNextItem(
    			ClientId,
    			Handle,
                CurrFileName.ItemId + 1
                );

            if ( dwErr == ERROR_NO_MORE_FILES )
                break;

            if ( dwErr != ERROR_SUCCESS )
            {
                LangErrorMessageBoxFmt( IDS_FUNC_FAILED_FMT, L"CFH_MoveToNextItem", dwErr );
                break;
            }

            continue;
        }

		if ( FileCreateResult != FILE_OP_CONTINUE )
		{
			DbgPrintf(
                "ProceedFileCreate failed(%u)\n",
                FileCreateResult
                );
            break;
		}

		if ( !CurrFileName.bHasStream )
		{
			if ( FileHandle != INVALID_HANDLE_VALUE )
			{
                SetFileTime( FileHandle, NULL, NULL, &CurrFileName.FileLastWriteTime );
				CloseHandle( FileHandle );

				FileHandle = INVALID_HANDLE_VALUE;
			}
			continue;
		}

        FILE_OP FileWriteOpRet = ProceedWriteStream(
            pCopyingHelper,
			ClientId,
			Handle,
            FileHandle,
			pFileBuffer,
			FileBufferSize,
            &FileCopyingProgressDlg,
            CurrentFileName,
            ProcessedItemsSize,
            StatusInfo,
            &DlgFileWriteCached,
            CurrFileName,
            &DlgFileReadFromHostCached
            );

        if ( FileHandle != INVALID_HANDLE_VALUE )
        {
            SetFileTime( FileHandle, NULL, NULL, &CurrFileName.FileLastWriteTime );
		    CloseHandle( FileHandle );
            FileHandle = INVALID_HANDLE_VALUE;
        }

        if ( FileWriteOpRet == FILE_OP_CANCEL_ALL )
            break;

        if ( FileWriteOpRet != FILE_OP_CONTINUE  )
        {
            //Skip 

            dwErr = pCopyingHelper->CFH_MoveToNextItem(
    			ClientId,
    			Handle,
                CurrFileName.ItemId + 1
                );

            if ( dwErr == ERROR_NO_MORE_FILES )
                break;

            if ( dwErr != ERROR_SUCCESS )
            {
                LangErrorMessageBoxFmt( IDS_FUNC_FAILED_FMT, L"CFH_MoveToNextItem(2)", dwErr );
                break;
            }

            continue;

        }

	}

    if ( pFileBuffer )
    {
        FREE( pFileBuffer );
        pFileBuffer = NULL;
    }

    //=======================

    FileCopyingProgressDlg.EndDialog();
    
    result = pCopyingHelper->CFH_End( ClientId, Handle );
	
	return ReturnResult;

}

FILE_OP
ProceedWriteStream(
    CFileCopierHelper *pCopyingHelper,
    IN  UINT ClientId,
	IN  UINT Handle,
    IN  OUT HANDLE &FileHandle,
    IN  PVOID pFileBuffer,
    IN  ULONG FileBufferSize,
    COM_CFileCopyingProgressDlg *pProgressDlg,
    IN  wstring & CurrentFileName,
    IN OUT ULONG64 &ProcessedItemsSize,
    IN COPY_FILES_BUILDING_LIST_STATUS &StatusInfo,
    IN int *pDlgFileWriteCached,
    IN COPY_FILES_GET_CURRENT_FILENAME &CurrFileName,
    IN OUT int *pDlgFileReadFromHostCached
    )
{
    DWORD dwErr = 0;
    BOOL bRet = FALSE;
    ULONG64 CurrentWrittenBytes = 0;

	while ( TRUE )
	{
		//CHAR szBuffer[1024];
		UINT ReadBytes = 0;
		UINT64 CurrOffset = 0;

repeat_read_file:

		dwErr = pCopyingHelper->CFH_GetCurrentStream(
			ClientId,
			Handle,
			pFileBuffer,
			FileBufferSize,
			&ReadBytes,
			&CurrOffset
			);

		if ( dwErr == ERROR_HANDLE_EOF ||
			 dwErr == ERROR_NO_DATA
			 )
		{
			break;
		}

        if ( dwErr != ERROR_SUCCESS && dwErr != ERROR_READ_FAULT )
        {
            wstring ClipboardFilePath = LangGetStringW( IDS_CLIPBOARDPATH_PREFIX );
            ClipboardFilePath += CurrFileName.wszRelFilePath;

            int FileReadFailDlgRet = DlgFileIoErrorCached(
                GetModuleInstance(),
                ClipboardFilePath.c_str(),
                LangGetStringW( IDS_CANNOT_READ_FILE ).c_str(),
                dwErr,
                pDlgFileReadFromHostCached
                );

            if ( FileReadFailDlgRet == DLG_SKIP || FileReadFailDlgRet == DLG_SKIP_ALL )
            {
                CloseHandle( FileHandle );
                FileHandle = INVALID_HANDLE_VALUE;
                DeleteFileW( CurrentFileName.c_str() );
                return FILE_OP_SKIP;
            }
            else if ( FileReadFailDlgRet == DLG_RETRY )
            {
                goto repeat_read_file;
            }
            else
            {
                CloseHandle( FileHandle );
                FileHandle = INVALID_HANDLE_VALUE;
                DeleteFileW( CurrentFileName.c_str() );
                return FILE_OP_CANCEL_ALL;
            }

        }

		if ( dwErr != ERROR_SUCCESS )
		{
			//DbgPrintf( "CFH_GetCurrentStream failed: %u(%ws)\n", dwErr, CurrFileName.wszRelFilePath );
            LangErrorMessageBoxFmt( IDS_FUNC_FAILED_FMT, L"CFH_GetCurrentStream", dwErr );

            CloseHandle( FileHandle );
            FileHandle = INVALID_HANDLE_VALUE;
            DeleteFileW( CurrentFileName.c_str() );
			return FILE_OP_CANCEL_ALL;
		}

        ProcessedItemsSize += ReadBytes;
        CurrentWrittenBytes += ReadBytes;

        pProgressDlg->ShowProgress(
            NULL,
            StatusInfo.FileCount + StatusInfo.FolderCount,
            ProcessedItemsSize,
            StatusInfo.AllStreamSize,
            CurrentFileName.c_str()
            );

        if ( pProgressDlg->IsCancelled() )
        {
            CloseHandle( FileHandle );
            FileHandle = INVALID_HANDLE_VALUE;
            DeleteFileW( CurrentFileName.c_str() );
			return FILE_OP_CANCEL_ALL;
        }

		DWORD dwWriteln = 0;
        DWORD dwNeedToWrite = ReadBytes;
        PUCHAR pszBuff = (PUCHAR)pFileBuffer;

write_retry:

		bRet = WriteFile(
			FileHandle,
			pszBuff,
			dwNeedToWrite,
			&dwWriteln,
			NULL
			);

        DWORD dwGle = GetLastError();

        if ( bRet && dwWriteln != dwNeedToWrite && dwWriteln != 0 )
        {
            dwNeedToWrite -= dwWriteln;
            pszBuff += dwWriteln;
            goto write_retry;
        }

		if ( !bRet )
		{
            int FileWriteFailDlgRet = DlgFileIoErrorCached(
                GetModuleInstance(),
                CurrentFileName.c_str(),
                LangGetStringW( IDS_CANNOT_WRITE_TO_FILE ).c_str(),
                dwGle,
                pDlgFileWriteCached
                );

            if ( FileWriteFailDlgRet == DLG_SKIP || FileWriteFailDlgRet == DLG_SKIP_ALL )
            {
                //Skip file

                CloseHandle( FileHandle );
                FileHandle = INVALID_HANDLE_VALUE;
                DeleteFileW( CurrentFileName.c_str() );

                ProcessedItemsSize -= CurrentWrittenBytes;
                ProcessedItemsSize += CurrFileName.FileSize;

                return FILE_OP_SKIP;
            }
            else if ( FileWriteFailDlgRet == DLG_RETRY )
            {
                goto write_retry;
            }
            else
            {
                return FILE_OP_CANCEL_ALL;
            }
		}
	} // while

    return FILE_OP_CONTINUE;
}
