/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "SharedMem.h"

CSharedMem::CSharedMem()
{
	m_hFileMap = NULL;
	m_pBase = NULL;
}

CSharedMem::~CSharedMem()
{
	Close();
}

VOID
CSharedMem::Close()
{
	if ( m_pBase != NULL )
	{
		UnmapViewOfFile( m_pBase );
		m_pBase = NULL;
	}

	if ( m_hFileMap != NULL )
	{
		CloseHandle( m_hFileMap );
		m_hFileMap = NULL;
	}
}


PVOID
CSharedMem::Create(
    LPCSTR lpszName,
    ULONG MemSize,
	PBOOL pExist
    )
{
	Close();

	SECURITY_ATTRIBUTES SecAttr;
	CHAR SecDesc[SECURITY_DESCRIPTOR_MIN_LENGTH];

    InitializeSecurityDescriptor( (PSECURITY_DESCRIPTOR)&SecDesc, SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl( (PSECURITY_DESCRIPTOR)&SecDesc, TRUE, 0, FALSE);

    SecAttr.nLength = sizeof(SecAttr);
    SecAttr.bInheritHandle = FALSE;
    SecAttr.lpSecurityDescriptor = (PSECURITY_DESCRIPTOR)&SecDesc;

	//Using NULL security descriptor

	m_hFileMap = CreateFileMappingA(
		INVALID_HANDLE_VALUE,
		&SecAttr,
		PAGE_READWRITE,
		0,
		MemSize,
		lpszName
		);

	DWORD dwGle = GetLastError();

	if ( NULL == m_hFileMap )
		return NULL;

	BOOL bExist = ( ERROR_ALREADY_EXISTS == dwGle );

	if (  pExist )
	{
		*pExist = bExist;
	}

	m_pBase = MapViewOfFile(
		m_hFileMap,
		FILE_MAP_READ | FILE_MAP_WRITE,
		0,0,
		MemSize
		);

	if ( NULL == m_pBase )
	{
		Close();
		return NULL;
	}

	//zero memory on init

	if ( !bExist )
	{
		RtlZeroMemory( m_pBase, MemSize );
	}


	return m_pBase;
}
